# _*_ coding: utf-8 _*_
"""
Time:     2021/12/2 22:04
Author:   ZHANG Yuwei
Version:  V 0.2
File:     lossFunc_sym.py
Describe:
"""
from scipy.special import gammaln
from numpy import log
import numpy as np


def LossFunctionConstantVariance(u, v, y, G, b1, b2, lambda2, weight, verbose=False):
    u = max(u, 0.01)
    v = max(v, 1e-09)
    fun1 = ((u ** 2) / v) * log(u)
    fun2 = -((u ** 2) * (1 / v) * log(v))
    fun3 = -gammaln(u ** 2 / v)
    fun4 = gammaln(y + u**2 / v)
    fun5 = -(y + u ** 2 / v) * log(1 + u / v)
    fun6 = - lambda2 * (np.dot(G[b1, :], G[b1, :].transpose()) + np.dot(G[b2, :], G[b2, :].transpose()))    # regularization
    fun = fun1 + fun2 + fun3 + fun4 + fun5 + fun6
    if verbose and b1 % 100 == 0 and b2 % 100 == 0:
        print(fun-fun6, fun6, fun6/lambda2, pow(y, weight))
    return fun * pow(y, weight)


def LossFunctionFano(u, b, y, G, b1, b2, lambda2, weight, verbose=False):
    u = max(u, 0.01)
    b = max(b, 1e-09)
    fun1 = - (u/b) * log(b)
    fun2 = -gammaln(u/b)
    fun3 = gammaln(y + (u/b))
    fun4 = -(y + (u/b))*log(1 + (1/b))
    fun5 = - lambda2 * (np.dot(G[b1, :], G[b1, :].transpose()) + np.dot(G[b2, :], G[b2, :].transpose()))
    fun = fun1 + fun2 + fun3 + fun4 + fun5
    if verbose and b1 % 100 == 0 and b2 % 100 == 0:
        print(fun-fun5, fun5, fun5/lambda2, pow(y, weight))
    # if b1==b2 == 464:
    #     print(u, b, y, np.round(fun1, 2), np.round(fun2,2), np.round(fun3,2), np.round(fun4,2), np.round(fun5,2), np.round(fun,2), b1)
    if np.isnan(fun) or np.isinf(fun):
        exit()
    return fun * pow(y, weight)


def LossFunctionConstantCoefficientVariation(u, a, y, G, b1, b2, lambda2, weight, verbose=False):
    u = max(u, 0.01)
    a = max(a, 1e-09)
    fun1 = -(1/a) * log(a)
    fun2 = -(1/a) * log(u)
    fun3 = -gammaln(1/a)
    fun4 = gammaln(y + 1/a)
    fun5 = -(y + 1/a) * log(1 + 1/(a * u))
    fun6 = - lambda2 * (np.dot(G[b1, :], G[b1, :].transpose()) + np.dot(G[b2, :], G[b2, :].transpose()))
    fun = fun1 + fun2 + fun3 + fun4 + fun5 + fun6
    if verbose and b1 % 100 == 0 and b2 % 100 == 0:
        print(fun-fun6, fun6, fun6/lambda2, pow(y, weight))
    return fun * pow(y, weight)



#! /usr/bin/env python
"""
knighRuizAlg is an implementation of the matrix balancing algorithm developed by Knight and Ruiz.
The goal is to take a matrix A and find a vector x such that, diag(x)*A*diag(x) returns a doubly stochastic matrix.
"""

import numpy as np
import matplotlib
matplotlib.use('Agg')


def get_matrix(matrix, result):
    new_matrix = np.copy(matrix)
    length = len(matrix)
    for i in range(length):
        for j in range(i, length, 1):
            if (not (np.isnan(result[i])) and not (np.isnan(result[j]))):
                new_matrix[i, j] = matrix[i, j] / (result[i] * result[j])
                new_matrix[j, i] = new_matrix[i, j]
    return new_matrix


class Compute_KR():
    def __init__(self, matrix, asy=False):
        self.matrix = matrix
        self.length = len(self.matrix)
        self.asy = asy
        self.valid_thre = None

    def run(self):
        norm = self.computeKR()
        factor = self.getSumFactor(norm)
        # print("factor", factor)
        factor = 1
        return np.asarray(norm) * factor

    def computeKR(self):
        recalculate = True
        offset = self.getOffset(0.0)
        iteration = 1
        max_iter = 60
        while (recalculate and iteration <= max_iter):
            sparsematrix = self.populateMatrix(self.matrix, offset)  # delete the rows/cols according to offset
            new_size = len(sparsematrix)
            # print("iteration = {}, total_bin = {}, left_bin = {}".format(iteration, self.length, new_size))
            x0 = [1.0] * new_size
            x0 = self.computeKRNormVector(sparsematrix, 1e-6, x0, 0.1)
            recalculate = False
            rowsTossed = 0
            if (x0 is None or iteration == max_iter-1):
                recalculate = True
                if (iteration < max_iter-1):    #x0 is None: not converse within 100 times
                    offset = self.getOffset(iteration)  # increase the threshold to delete more data
                else:   #iteration==max_iter-1
                    offset = self.getOffset(max_iter+5)
                    # print(iteration, "Changed threshold to 25%.")
            else:   #x0 is valid
                kr = np.zeros(self.length)
                for i in range(self.length):
                    if (offset[i] == -1):
                        kr[i] = np.nan
                    else:
                        kr[i] = 1.0 / x0[offset[i]]  # transform the factor
                mySum = self.getSumFactor(kr)  # an overall factor to the whole matrix
                index = 0
                if self.valid_thre == None:
                    if self.asy == True:
                        test = kr * mySum
                        self.valid_thre = np.min(test[np.where(np.isnan(test)==False)])
                    else:
                        self.valid_thre = 0.01
                # print(self.valid_thre, "valid_threshold")
                for i in range(self.length):
                    if kr[i] * mySum < self.valid_thre:  # factor is valid but too little
                        offset[i] = -1  # delete the row and calculate again
                        rowsTossed += 1
                        recalculate = True
                    elif (offset[i] != -1):  # valid
                        offset[i] = index
                        index += 1
            # print("Finish Iteration = {}, recalculate: {}".format(iteration, recalculate))
            iteration += 1
        if (iteration > max_iter and recalculate):
            # print("Iteration > {}, output all nan.".format(max_iter))
            kr = [np.nan] * self.length
        return kr

    def getOffset(self, percent=0.0):
        """
        set rows/cols that cannot reach the threshold to -1 in offset list
        :param matrix: input matrix
        :return: offset list (all deleted rows are set to -1)
        """
        rowSums = np.zeros(self.length)
        for i in range(self.length):
            rowSums[i] = np.sum(self.matrix[i,])
        thresh = 0.0
        if (percent != 0.0):
            no_zeros = np.where(rowSums != 0)
            posRowSums = np.sort(rowSums[no_zeros])
            thresh = posRowSums[int(len(posRowSums) * percent / 100)]
        # print(thresh)
        offset = [0] * self.length
        index = 0
        for i in range(self.length):
            if (np.sum(self.matrix[i,]) <= thresh):
                offset[i] = -1
            else:
                offset[i] = index
                index += 1
        return offset

    def populateMatrix(self, matrix, offset):
        dele_row = np.where(np.asarray(offset) == -1)[0]
        matrix = np.delete(matrix, dele_row, axis=0)
        matrix = np.delete(matrix, dele_row, axis=1)
        return matrix

    def computeKRNormVector(self, matrix, tol, x0, delta=0.1, Delta=3):
        """
            A balancing algorithm for symmetric matrics, which attempts to find a vector x0 such that
            diag(x0)*A*diag(x0) is close to doubly stochastic.

            :param matrix: input matrix, must be symmetric and nonnegative.
            :param tol: error tolerance
            :param x0: initial guess, default: ones
            :param delta/Delta: how close/far balancing vectors can get to/from the edge of the positive cone.
            :return: balancing vector x0.
	    """
        n = len(x0)  # No. selected rows
        e = [1.0] * n   #initialization
        # Inner stopping criterion parameters
        g = 0.9
        etamax = 0.1
        eta = etamax
        rt = pow(tol, 2)  # default: 1e-12
        v = x0 * (matrix.dot(x0))  # shape = (509,)
        rk = 1.0 - v
        rho_km1 = rk.transpose().dot(rk)  # a scaler
        rout = rold = rho_km1
        MVP = 0  # We will count matrix vector products
        not_changing = 0  # Outer iteration count
        while (rout > rt and not_changing < 100):
            k = 0
            y = np.copy(e)
            rho_km2 = rho_km1
            innertol = max(pow(eta, 2) * rout, rt)
            while (rho_km1 > innertol):  # inner iteration by CG
                k += 1
                if (k == 1):
                    Z = rk / v
                    p = np.copy(Z)
                    rho_km1 = rk.transpose().dot(Z)
                else:
                    beta = rho_km1 / rho_km2
                    p = Z + beta * p
                if k > 10:
                    break
                # update search direction efficiently.
                tmp = matrix.dot(x0 * p)
                w = x0 * tmp + v * p
                alpha = rho_km1 / (p.transpose().dot(w))
                ynew = y + alpha * p
                # test distance to boundary of cone
                minynew = np.amin(ynew)
                if (minynew <= delta):
                    if (delta == 0):
                        break
                    gamma = np.inf
                    for i in range(len(ynew)):
                        if (alpha * p[i] < 0 and (delta - y[i]) / (alpha * p[i]) < gamma):
                            gamma = (delta - y[i]) / (alpha * p[i])
                    y += gamma * alpha * p
                    break
                maxynew = np.amax(ynew)
                if (maxynew >= Delta):
                    gamma = np.inf
                    for i in range(len(ynew)):
                        if (ynew[i] > Delta and (Delta - y[i]) / (alpha * p[i]) < gamma):
                            gamma = (Delta - y[i]) / (alpha * p[i])
                    y += gamma * alpha * p
                    break
                rho_km2 = rho_km1
                y = np.copy(ynew)
                rk -= alpha * w
                Z = rk / v
                rho_km1 = rk.transpose().dot(Z)
                # print("-----------------")
            x0 = x0 * y
            v = x0 * matrix.dot(x0)
            rk = 1.0 - v
            rho_km1 = rk.transpose().dot(rk)
            # print(rho_km1, rout, abs(rho_km1 - rout))
            if (abs(rho_km1 - rout) < 1e-6 or np.isinf(rho_km1)):
                not_changing += 1
            rout = rho_km1
            MVP += k + 1
            # Update inner iteration stopping criterion
            rat = rout / rold
            rold = rout
            r_norm = np.sqrt(rout)
            eta_o = eta
            eta = g * rat
            if (g * pow(eta_o, 2) > 0.1):
                eta = max(eta, g * pow(eta_o, 2))
            eta = max(min(eta, etamax), 0.5 * tol / r_norm)
            # print("not_changing time = " + str(not_changing))
        if (not_changing >= 100):
            return None
        return x0

    def getSumFactor(self, norm_list):
        """
		Use the calculated bin factors to calculate the overall factor to the whole normalized matrix.
	    :param norm_list: the calculated factor list
	    :return: an overall factor to the whole matrix
	    """
        matrix_sum = 0  # raw pixel sum
        norm_sum = 0  # normalized pixel sum
        for i in range(self.length):
            for j in range(i, self.length, 1):
                if (not (np.isnan(norm_list[i])) and not (np.isnan(norm_list[j])) and norm_list[i] > 0 and norm_list[
                    j] > 0):  # only if the factor is larger than 0, valid
                    if (i == j):  # pixel on the diagonal
                        norm_sum += self.matrix[i, j] / (norm_list[i] * norm_list[j])
                        matrix_sum += self.matrix[i, j]
                    else:
                        norm_sum += (2 * self.matrix[i, j]) / (norm_list[i] * norm_list[j])
                        matrix_sum += 2 * self.matrix[i, j]
        return np.sqrt(norm_sum / matrix_sum)


def KR_norm(matrix_in):

    result = Compute_KR(matrix_in).run()
    matrix_out = get_matrix(matrix_in, result)

    return matrix_out





# _*_ coding: utf-8 _*_
"""
Time:     2021/7/22 17:36
Author:   WANG Bingchen
Version:  V 0.1
File:     initialParams.py
Describe: 
"""
import numpy as np
from scipy.optimize import minimize
from scipy.special import gammaln
from numpy import log
import warnings


def initialMatrices(U_0, K):
    U = U_0
    u, sigma, vt = np.linalg.svd(U)
    print(sigma[:10])
    u = u[:, :K]
    sigma = sigma[:K]
    vt = vt[:K, :]
    G = np.dot(u, np.diag(sigma))    # U * sigma
    H = vt # V.T

    return G, H

def initialMatrix_sym(U_0, K):
    U = U_0
    values, vectors = np.linalg.eigh(U) # increasing order
    positives = len(np.where(values>0)[0])
    if K > positives:
        # values = np.abs(values)
        K = positives
        print("Change the dimension to {}".format(K))
    order = np.arange(-1, -K - 1, -1)   # take the largest K eigenvalue
    values = values[order]
    vectors = vectors[:, order]
    return np.dot(vectors, np.diag(np.sqrt(values)))    # X where U = X * X.T

def loglikv(args):
    mu, y = args
    fun = lambda x: -(sum((mu**2)*log(mu) / x[0]) -
                      sum((mu**2)*log(x[0]) / x[0]) -
                      sum(gammaln(mu**2 / x[0])) +
                      sum(gammaln(y + mu**2 / x[0])) -
                      sum((y + mu**2 / x[0])*log(1 + mu / x[0])))
    return fun

def loglikb(args):
    mu, y = args
    fun = lambda x: -(sum((mu / x[0]) * log(1 / x[0])) -
                      sum(gammaln(mu / x[0])) +
                      sum(gammaln(y + (mu / x[0]))) -
                      sum((y + (mu / x[0])) * log(1 + (1 / x[0]))))
    return fun


def loglika(args):
    mu, y = args
    n = len(y)
    fun = lambda x: -(n / x[0]*log(1 / x[0]) -
                      sum(1 / x[0]*log(mu)) -
                      n*gammaln(1 / x[0]) +
                      sum(gammaln(y + 1 / x[0])) -
                      sum((y + 1 / x[0])*log(1 + 1 / (x[0]*mu))))
    return fun


def con(args):
    Xmin, Xmax = args
    cons = ({'type': 'ineq', 'fun': lambda x: x[0] - Xmin},
            {'type': 'ineq', 'fun': lambda x: Xmax - x[0]})
    return cons


def getv(dataFrame, mu):
    warnings.filterwarnings("ignore")
    (genes, cells) = dataFrame.shape
    vg = []
    for g in range(genes):
        mu_g = mu[g, :]
        Y_g = np.array(dataFrame.iloc[g])
        x0 = np.asarray((1))
        args = (mu_g, Y_g)
        args1 = (1e-07, 2)
        cons = con(args1)
        res = minimize(loglikv(args), x0=x0, constraints=cons)
        if res.success:
            vg.append(res.x[0])
        else:
            vg.append(1)
    vg = np.array(vg)
    vg = vg.reshape(genes, 1)
    return vg


def getb(dataFrame, mu):
    warnings.filterwarnings("ignore")
    (genes, cells) = dataFrame.shape
    bg = []
    for g in range(genes):
        mu_g = mu[g, :]
        Y_g = np.array(dataFrame.iloc[g])
        x0 = np.asarray((1))
        args = (mu_g, Y_g)
        args1 = (1e-07, 2)
        cons = con(args1)
        res = minimize(loglikb(args), x0=x0, constraints=cons)
        if res.success:
            bg.append(res.x[0])
        else:
            bg.append(1)
    bg = np.array(bg)
    bg = bg.reshape(genes, 1)
    return bg


def geta(dataFrame, mu):
    warnings.filterwarnings("ignore")
    (genes, cells) = dataFrame.shape
    ag = []
    for g in range(genes):
        mu_g = mu[g, :]
        Y_g = np.array(dataFrame.iloc[g])
        x0 = np.asarray((1))
        args = (mu_g, Y_g)
        args1 = (1e-07, 2)
        cons = con(args1)
        res = minimize(loglika(args), x0=x0, constraints=cons)
        if res.success:
            ag.append(res.x[0])
        else:
            ag.append(1)
    ag = np.array(ag)
    ag = ag.reshape(genes, 1)
    return ag


# _*_ coding: utf-8 _*_
"""
Time:     2021/7/22 17:33
Author:   WANG Bingchen
Version:  V 0.1
File:     lossFunc.py
Describe:
"""
from scipy.special import gammaln
from numpy import log
import numpy as np


def LossFunctionConstantVariance(u, v, y, G, H, g, c, lambda2, weight, verbose=False):  # maximize the obj
    u = max(u, 0.01)
    v = max(v, 1e-09)
    fun1 = ((u ** 2) / v) * log(u)
    fun2 = -(u ** 2 * (1 / v) * log(v))
    fun3 = -gammaln(u ** 2 / v)
    fun4 = gammaln(y + u**2 / v)
    fun5 = -(y + u ** 2 / v) * log(1 + u / v)
    fun6 = - lambda2 * (np.dot(G[g, :], G[g, :].transpose()) + np.dot(H[:, c].transpose(), H[:, c]))
    fun = fun1 + fun2 + fun3 + fun4 + fun5 + fun6
    if verbose and g % 100 == 0 and c % 100 == 0:
        print(fun-fun6, fun6, fun6/lambda2, pow(y, weight))
    return fun * pow(y, weight)


def LossFunctionFano(u, b, y, G, H, g, c, lambda2, weight, verbose=False):
    u = max(u, 0.01)
    b = max(b, 1e-09)
    fun1 = - (u/b) * log(b)
    fun2 = -gammaln(u/b)
    fun3 = gammaln(y + (u/b))
    fun4 = -(y + (u/b))*log(1 + (1/b))
    fun5 = - lambda2 * (np.dot(G[g, :], G[g, :].transpose()) + np.dot(H[:, c].transpose(), H[:, c]))
    fun = fun1 + fun2 + fun3 + fun4 + fun5
    if verbose and g % 100 == 0 and c % 100 == 0:
        print(fun-fun5, fun5, fun5/lambda2, pow(y, weight))
    return fun * pow(y, weight)


def LossFunctionConstantCoefficientVariation(u, a, y, G, H, g, c, lambda2, weight, verbose=False):
    u = max(u, 0.01)
    a = max(a, 1e-09)
    fun1 = -(1/a) * log(a)
    fun2 = -(1/a) * log(u)
    fun3 = -gammaln(1/a)
    fun4 = gammaln(y + 1/a)
    fun5 = -(y + 1/a) * log(1 + 1/(a * u))
    fun6 = - lambda2 * (np.dot(G[g, :], G[g, :].transpose()) + np.dot(H[:, c].transpose(), H[:, c]))
    fun = fun1 + fun2 + fun3 + fun4 + fun5 + fun6
    if verbose and g % 100 == 0 and c % 100 == 0:
        print(fun-fun6, fun6, fun6/lambda2, pow(y, weight))
    return fun * pow(y, weight)



import numpy as np
from math import log2


def get_v(M, sparse_m, vs):
    # intra- and inter- affinity
    # return np.sum(M[vs, :]) - np.sum([M[i, i] for i in vs])
    v = np.sum(M[vs, :])
    if v == 0:
        v += 1
    return v
    # return sparse_m[vs, :].sum()  # slow


def get_s(M, sparse_m, vs):
    # intra-affinity
    return np.sum(M[np.ix_(vs, vs)])
    ''' slow
    c = 0
    for i in vs:
        for j in vs:
            c += M[i, j]
    return c
    '''


def get_g(M, sparse_m, vs):
    # inter-affinity
    return get_v(M, sparse_m, vs) - get_s(M, sparse_m, vs)


def get_se(vG, g, V, pV):
    return float(g)/float(vG)*log2(float(pV)/float(V))


def get_se_for_vertices(M, sparse_m, vG, vs, pV):
    g = get_g(M, sparse_m, vs)
    V = get_v(M, sparse_m, vs)
    return get_se(vG, sparse_m, g, V, pV)


def get_se_for_vertex(M, sparse_m, d, vG, bin, pV):
    if get_se(vG, d[bin], d[bin], pV) != get_se_for_vertices(M, vG, [bin], pV):
        raise ValueError("bin se error")
    return get_se_for_vertex(M, sparse_m, vG, [bin], pV)


def get_delta_merge_se(M, sparse_m, vG, d, parent, node1, node2):
    # leaf merge
    new_V = node1.V + node2.V
    new_s = node1.s + node2.s
    new_s_tmp = 0
    for b1 in node1.vs:
        for b2 in node2.vs:
            new_s_tmp += M[b1, b2]
    # new_s_tmp = np.sum(M[np.ix_(node1.vs, node2.vs)])   # slow
    new_s += 2*new_s_tmp
    new_se = -(new_V-new_s)/vG*log2(new_V/vG)
    new_se += (new_V*log2(new_V) - node1.d_log_d - node2.d_log_d) / vG
    old_se = node1.se + node2.se \
         + (node1.V*node1.log_V - node1.d_log_d + node2.V*node2.log_V - node2.d_log_d)/vG
    delta = old_se - new_se
    return delta


def get_delta_combine_se(M, sparse_m, vG, parent, node1, node2):
    new_node_V = node1.V + node2.V
    new_node_s = node1.s + node2.s
    new_s_tmp = 0
    for b1 in node1.vs:
        for b2 in node2.vs:
            new_s_tmp += M[b1, b2]
    # new_s_tmp = np.sum(M[np.ix_(node1.vs, node2.vs)])  # slow
    new_node_s += 2*new_s_tmp
    new_node_g = new_node_V - new_node_s
    new_node_se = get_se(vG, new_node_g, new_node_V, parent.V)
    new_node1_se = get_se(vG, node1.g, node1.V, new_node_V)
    new_node2_se = get_se(vG, node2.g, node2.V, new_node_V)
    return node1.se + node2.se - new_node_se - new_node1_se - new_node2_se


def get_delta_nm(M, sparse_m, vG, parent, node1, node2, r=1):
    if len(node1.vs) == 1:
        node1_q = - r*np.power(node1.V/vG, 2)
        # node1_q = node1.g/vG - r*np.power(node1.V/vG, 2)
    else:
        node1_q = node1.s/vG - r*np.power(node1.V/vG, 2)
    if len(node2.vs) == 1:
        node2_q = - r*np.power(node2.V/vG, 2)
        # node2_q = node2.g/vG - r*np.power(node2.V/vG, 2)
    else:
        node2_q = node2.s/vG - r*np.power(node2.V/vG, 2)
    new_node_V = node1.V + node2.V
    new_node_s = node1.s + node2.s
    new_s_tmp = 0
    for b1 in node1.vs:
        for b2 in node2.vs:
            new_s_tmp += M[b1, b2]
    new_node_s += 2*new_s_tmp
    # new_node_g = new_node_V - new_node_s

    after = new_node_s/vG - r*np.power(new_node_V/vG, 2)
    return after - (node1_q + node2_q)



from sklearn.neighbors import kneighbors_graph
import numpy as np


def build_graph_ST(gene_m, coord_m):
    gene_m = gene_m.to_numpy()
    coord_m = coord_m.to_numpy()

    #gene_g = kneighbors_graph(gene_m, 10) + 1
    #coord_g = kneighbors_graph(coord_m, 10) + 1
    gene_g = np.matmul(gene_m, gene_m.T) + 1
    coord_g = np.matmul(coord_m, coord_m.T) + 1
    print(np.log(gene_g/coord_g))

    return gene_g*np.log(gene_g/coord_g)



from os import confstr_names
import numpy as np
from math import log2
from scipy.cluster import hierarchy
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph
from scipy import sparse
import heapq
import itertools
import networkx as nx
from itertools import chain
import pandas as pd
from queue import Queue
import kmeans1d
import scipy.linalg as la
from scipy.sparse import csgraph
# import umap
import time

import seat_wrapper


class SEAT(AgglomerativeClustering):

    def __init__(self, min_k=1, max_k=4,
                 max_g_ratio=0,
                 affinity='precomputed',
                 strategy='top_down',
                 objective='structure_entropy',
                 n_neighbors=10,
                 corr_cut_off=0.8,
                 ):
        self.min_k = min_k
        self.max_k = max_k
        self.ks = range(min_k, max_k+1)
        self.max_g_ratio = max_g_ratio
        self.affinity = affinity
        self.strategy = strategy
        self.objective = objective
        self.n_neighbors = n_neighbors
        self.corr_cut_off = corr_cut_off

    def get_affinity(self, X):
        if self.affinity == 'precomputed':
            aff_m = X
        elif self.affinity == 'knn_neighbors':
            # X = umap.UMAP().fit_transform(np.matrix(X))
            aff_m = kneighbors_graph(X, self.n_neighbors).toarray()
            aff_m = (aff_m + aff_m.T)/2
            aff_m[np.nonzero(aff_m)] = 1

        elif self.affinity == 'T10':
            # idx = np.argwhere(np.std(X, axis=1) == 0).T[0]
            # X[idx, 0] = X[idx, 0] + 0.0001
            # aff_m = np.corrcoef(X)
            from scipy.stats import spearmanr
            from scipy.spatial.distance import squareform, pdist
            aff_m, _ = spearmanr(X.T)
            aff_m[aff_m < 0] = 0
            dist_m = squareform(pdist(X, metric='euclidean'))
            sigma = 10
            aff_m2 = np.exp(-dist_m*dist_m/(2*sigma*sigma))
            aff_m2[aff_m2 < 0] = 0
            aff_m = (0.8*aff_m + 0.2*aff_m2)/2
            aff_m[aff_m < 0.25] = 0

        elif self.affinity == 'T16':
            # idx = np.argwhere(np.std(X, axis=1) == 0).T[0]
            # X[idx, 0] = X[idx, 0] + 0.0001
            # aff_m = np.corrcoef(X)
            from scipy.stats import spearmanr
            from scipy.spatial.distance import squareform, pdist
            aff_m, _ = spearmanr(X.T)
            aff_m[aff_m < 0] = 0
            dist_m = squareform(pdist(X, metric='euclidean'))
            sigma = 10
            aff_m2 = np.exp(-dist_m*dist_m/(2*sigma*sigma))
            aff_m2[aff_m2 < 0] = 0
            aff_m = (0.8*aff_m + 0.2*aff_m2)/2
            # aff_m[aff_m < 0.25] = 0

        self.affinity_m = aff_m

    def fit(self, X, y=None):

        X = self._validate_data(X, ensure_min_samples=2, estimator=self)

        if self.min_k is not None and self.min_k <= 0:
            raise ValueError("min_k should be an integer greater than 0."
                             " %s was provided." % str(self.min_k))

        if self.max_k is not None and self.max_k <= 2:
            raise ValueError("max_k should be an integer greater than 2."
                             " %s was provided." % str(self.max_k))

        if self.affinity not in ['precomputed', 'knn_neighbors', 'T10', 'T16']:
            raise ValueError("affinity should be precomputed, knn_neighbors, correlation."
                             " %s was provided." % str(self.affinity))

        if self.strategy not in ['bottom_up', 'top_down']:
            raise ValueError("affinity should be bottom_up, top_down."
                             " %s was provided." % str(self.strategy))

        self.get_affinity(X)

        # build the tree

        setree_class = seat_wrapper.SETree

        se_tree = setree_class(self.affinity_m, self.min_k, self.max_k,
                        self.max_g_ratio,
                        self.objective,
                        self.strategy)
        self.se_tree = se_tree
        t1 = time.time()
        Z = se_tree.build_tree()
        t2 = time.time()
        #print(Z[:20, :])
        print('build tree time', t2 - t1)
        t3 = time.time()
        se_tree.cut_tree(Z, self.ks)
        t4 = time.time()
        print('cut tree time', t4 - t3)
        print('build + cut tree time', t4 - t1)

        self.affinity_m = se_tree.get_affinity_m()
        self.vertex_num = se_tree.get_vertex_num()
        self.cost_m = se_tree.get_cost_m()
        self.cutoff_m = se_tree.get_cutoff_m()
        self.cluster_m = se_tree.get_cluster_m()
        self.optimal_k = se_tree.get_optimal_k()
        self.optimal_se = se_tree.get_optimal_se()
        self.labels_ = se_tree.get_optimal_cluster()
        print(self.labels_)
        self.submodules = se_tree.get_submodules()

        tmp = pd.DataFrame(np.matrix(self.cost_m[-1, 1:]), columns=self.ks).T # ???
        tmp['K'] = tmp.index
        tmp.columns = ['SE Score', 'K']
        self.se_scores = tmp
        self.delta_se_scores = self.se_scores[1:] - self.se_scores[:-1]

        self.Z_ = Z[:, :4]
        self.leaves_list = hierarchy.leaves_list(self.Z_)
        self.order = self._order()
        self.ks_clusters = pd.DataFrame(self.cluster_m, columns=['K={}'.format(k) for k in self.ks])
        Z_clusters = hierarchy.cut_tree(Z[:, :4], n_clusters=self.ks)
        self.Z_clusters = pd.DataFrame(np.matrix(Z_clusters), columns=['K={}'.format(k) for k in self.ks])

        self.submodule_k = len(set(self.submodules))

        self.newick = se_tree.to_newick()

        return self

    def _order(self):
        order = [(l, i) for i, l in enumerate(self.leaves_list)]
        order.sort()
        return [i for l, i in order]

    def _get_submodules(self):
        leafs = sorted([(self.order[l.vs[0]], l.id) for l in self.se_tree.leafs])
        order = [(v, i) for i, l in enumerate(leafs) for v in self.se_tree.node_list[l[1]].vs]
        order.sort()
        return [i for n, i in order]

    def oval_embedding(self, a=3, b=2, k=0.2):
        angle = np.array([self.order])*(2*np.pi/len(self.order))
        xcor = a*np.cos(angle)
        ycor = b*np.sqrt(np.exp(k*a*np.cos(angle)))*np.sin(angle)
        plane_coordinate = np.concatenate((xcor, ycor), axis=0).T
        return plane_coordinate



import umap
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator


class Node():

    def __init__(self, id, depth, label,
                 parent=None,
                 children=None,
                 child_ls=None,
                 vs=None,
                 ):
        self.id = id
        self.depth = depth
        self.label = label
        self.parent = parent
        self.children = children or []
        self.child_ls = child_ls or []
        self.vs = vs or []


class Tree():

    def __init__(self, y):
        self.y = y
        self.node_id = 0

        self.root = Node(self.update_node_id(), 0, 1)
        self.node_list = {}
        self.node_list[self.root.id] = self.root
        self._parse_y(y)

    def update_node_id(self, increment=True):
        if increment:
            self.node_id += 1
        else:
            self.node_id -= 1
        return self.node_id

    def get_current_node_id(self):
        return self.node_id

    def _parse_y(self, y):
        self._parse_y_aux(y, self.root)

    def _parse_y_aux(self, y, parent):
        if y.shape[1] == 1:
            return
        for l in set(y[:, 0]):
            vs = np.where(y[:, 0]==l)[0]
            node = Node(self.update_node_id(), parent.depth+1, l, parent=parent)
            node.vs = vs
            parent.children.append(node)
            parent.child_ls.append(l)
            self._parse_y_aux(y[vs, 1:], node)


class HierarchicalEmbedding(BaseEstimator):

    def __init__(self,
                 strategy='umap',
                 n_neighbors=10,
                 rand_coord_std=0.2,
                 ):
        self.strategy = strategy
        self.n_neighbors = n_neighbors
        self.rand_coord_std = rand_coord_std

    def fit(self, X, y):
        X = self._validate_data(X, ensure_min_samples=2, estimator=self)

        if self.strategy not in ['umap', 'spectral_embedding']:
            raise ValueError("affinity should be umap, spectral_embedding."
                             " %s was provided." % str(self.strategy))

        tree = Tree(y)
        self.look_tree(tree.root)
        # print(self.assign_rand_coords(y[:, 0]))

    def look_tree(self, node):
        if node.child_ls:
            print(node.id, node.depth, node.label)
            print(node.child_ls)
            for child in node.children:
                self.look_tree(child)

    def assign_rand_coords(self, labels):
        n = int(np.ceil(np.sqrt(len(set(labels)))))
        coords = np.zeros((len(labels), 2))
        coords[:, 0] = labels / n + np.random.normal(0, self.rand_coord_std, len(labels))
        coords[:, 1] = labels % n + np.random.normal(0, self.rand_coord_std, len(labels))

        return coords

    def embed_with_coords(self, X):
        mapper = umap.UMAP(
            random_state=0,
            n_neighbors=10,
            metric='precomputed',
            target_metric='l1',
            min_dist=0.01)

        embed = mapper.fit_transform(X)
        meta_df[['x','y']] = embed
        meta_df[['sub_x','sub_y']] = embed


import matplotlib.pyplot as plt
aff_m = np.genfromtxt('../../data/cellbench/p3cl_SEAT-A_affinity_m.csv', delimiter=',')
meta_df = pd.read_csv('../../data/cellbench/p3cl_cluster_result.csv')

he = HierarchicalEmbedding()
he.fit(aff_m, meta_df[['SEAT-D', 'SEAT-D(club)', 'SEAT-D(order)']].to_numpy())





# _*_ coding: utf-8 _*_
"""
Time:     2021/7/22 17:31
Author:   WANG Bingchen
Version:  V 0.1
File:     utiles.py
Describe:
"""

import numbers


def check_positive(**params):
    """Check that parameters are positive as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] <= 0:
            raise ValueError(
                "Expected {} > 0, got {}".format(p, params[p]))


def check_int(**params):
    """Check that parameters are integers as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if not isinstance(params[p], numbers.Integral):
            raise ValueError(
                "Expected {} integer, got {}".format(p, params[p]))


def check_bool(**params):
    """Check that parameters are bools as expected

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] is not True and params[p] is not False:
            raise ValueError(
                "Expected {} boolean, got {}".format(p, params[p]))


def check_between(v_min, v_max, **params):
    """Checks parameters are in a specified range

    Parameters
    ----------

    v_min : float, minimum allowed value (inclusive)

    v_max : float, maximum allowed value (inclusive)

    params : object
        Named arguments, parameters to be checked

    Raises
    ------
    ValueError : unacceptable choice of parameters
    """
    for p in params:
        if params[p] < v_min or params[p] > v_max:
            raise ValueError("Expected {} between {} and {}, "
                             "got {}".format(p, v_min, v_max, params[p]))


def check_noise_model(**params):
    for p in params:
        if (params[p] != "CV") and (params[p] != "Fano") and (params[p] != "CCV"):
            raise ValueError("Expected {} be one in CV, Fano, or CCV, got {}".format(p, params[p]))



def dataNormalization(dataFrame):
    """
    Normalization based on library size of each cell/bin.
    Each bin share the same coverage.

    Parameters:

    -----------

    :param dataFrame: origin dataframe

    -----------

    :return: dataframe normalized，sizefactors
    """
    Ngene, Ncell = dataFrame.shape
    librarySize = dataFrame.sum(axis=0) # sum of a column (for a cell)
    meanLibrarySize = librarySize.mean()
    sizeFactors = librarySize / meanLibrarySize # (Ncell, )
    for i in range(Ncell):
        if sizeFactors[i]:
            dataFrame.iloc[:, i] = dataFrame.iloc[:, i] / sizeFactors[i]
    #normalizedDataFrame = dataFrame / sizeFactors
    return dataFrame, sizeFactors





# _*_ coding: utf-8 _*_
"""
Time:     2021/7/22 17:29
Author:   WANG Bingchen
Version:  V 0.1
File:     __init__.py
Describe: 
"""
from .version import __version__
from .model import SMURF
from .model_sym import SMURF as SMURF_SYM

name = 'smurf-imputation'



# _*_ coding: utf-8 _*_
"""
Time:     2021/12/2 21:47
Author:   ZHANG Yuwei
Version:  V 0.2
File:     model_sym.py
Describe: SMURF for symmetric matrix
"""

from .lossFunc_sym import *
from .initialParams import *
import pandas as pd
from .optimize import CVOptimize_sym, FanoOptimize_sym, CCVOptimize_sym
from . import utiles
from .cell_circle import CellCircle
import warnings

warnings.filterwarnings("ignore")


class SMURF():
    def __init__(self, n_features=None, steps=1000, alpha=1e-5, eps=1e-3, lambda2=0.1, noise_model="Fano",
                 calculateIntialNoiseFactor=False, estimate_only=False, weight=0., verbose=False):

        self.K = n_features
        if self.K:
            self.batchSize = n_features * 10
        else:
            self.batchSize = None
        self.steps = steps
        self.alpha = alpha
        self.eps = eps
        self.lambda2 = lambda2
        self.noise_model = noise_model
        self.calculateLossFunc = True
        self.estmate_only = estimate_only
        self.calculateIntialNoiseFactor = calculateIntialNoiseFactor
        self.dataweightControl = weight  # add a controller on the data weighted gradient, y ** weight
        self.verbose = verbose

    def _check_params(self):
        """Check parameters

        This allows us to fail early - otherwise certain unacceptable
        parameter choices, such as n='10.5', would only fail after
        minutes of runtime.

        Raises
        ------
        ValueError : unacceptable choice of parameters
        """
        utiles.check_positive(n_features=self.K, eps=self.eps, batchsize=self.batchSize, alpha=self.alpha,
                              steps=self.steps)
        utiles.check_int(n_features=self.K, steps=self.steps)
        utiles.check_between(v_min=0, v_max=min(self.genes, self.cells), n_features=self.K)
        utiles.check_bool(iteration=self.calculateIntialNoiseFactor)
        utiles.check_noise_model(noise_model=self.noise_model)

    def MFConstantVariance(self):   # constance variance for each bin, v_b
        G = self.G.copy()
        A = self.A.copy()
        u = np.dot(G, G.T)

        if self.calculateIntialNoiseFactor:
            Vb = getv(self.A, u)    # (Nbin, 1)
            v = np.sqrt(np.dot(Vb, Vb.transpose()))  # v = sqrt(vi*vj)
        else:
            Vb = np.ones((self.bins, 1))
            v = np.sqrt(np.dot(Vb, Vb.transpose()))

        LossFuncGre = 0
        LossFunc = 0
        Lossf = []

        nonZerosInd = np.nonzero(np.triu(A))
        numNonZeros = nonZerosInd[0].size
        nonZeroElem = np.concatenate((nonZerosInd[0].reshape(numNonZeros, 1), nonZerosInd[1].reshape(numNonZeros, 1)),
                                     axis=1)

        for step in range(self.steps):

            if self.calculateLossFunc:
                for element in nonZeroElem:
                    b1 = element[0]
                    b2 = element[1]

                    LossFuncCG = LossFunctionConstantVariance(u[b1][b2], v[b1][b2], A[b1][b2], G, b1, b2, self.lambda2, self.dataweightControl, self.verbose)
                    LossFunc = LossFunc + LossFuncCG
            else:
                LossFunc = (step + 1) * (self.eps + 1)

            if abs(LossFunc - LossFuncGre) < self.eps:
                Lossf.append(LossFunc)
                # print("already converge")
                break
            else:
                Lossf.append(LossFunc)
                LossFuncGre = LossFunc
                LossFunc = 0

                np.random.shuffle(nonZeroElem)

                batchElements = nonZeroElem[0:self.batchSize - 1, :]

                for element in batchElements:
                    b1 = element[0]
                    b2 = element[1]
                    if u[b1][b2] <= 0.01:
                        u[b1][b2] = 0.01
                    if Vb[b1][0] <= 1e-09:
                        Vb[b1][0] = 1e-09
                    if Vb[b2][0] <= 1e-09:
                        Vb[b2][0] = 1e-09
                    v[b1][b2] = np.sqrt(Vb[b1][0]*Vb[b2][0])

                    dG1, dG2, dV1, dV2 = CVOptimize_sym(A, G, u, Vb, v, b1, b2, self.lambda2, self.dataweightControl)
                    G[b1, :] = G[b1, :] + self.alpha * dG1
                    G[b2, :] = G[b2, :] + self.alpha * dG2
                    Vb[b1, :] = Vb[b1, :] + self.alpha * dV1
                    Vb[b2, :] = Vb[b2, :] + self.alpha * dV2

                u = np.dot(G, G.T)
                v = np.sqrt(np.dot(Vb, Vb.transpose()))
            print("number of iteration: ", step + 1, "/", self.steps, LossFuncGre)
        u = np.dot(G, G.T)
        u[u < 0] = 0
        return u, G

    def MFFano(self):
        G = self.G.copy()
        A = self.A.copy()
        u = np.dot(G, G.T)
        alpha = self.alpha

        if self.calculateIntialNoiseFactor:
            bg = getb(self.A, u)    # (Nbin, 1)
        else:
            bg = np.ones((self.bins, 1)) * 1.0

        b = np.sqrt(np.dot(bg, bg.transpose())) # (Nbin, Nbin)

        LossFunc = 0
        LossFuncGre = 0
        Lossf = []

        nonZerosInd = np.nonzero(np.triu(A))
        numNonZeros = nonZerosInd[0].size
        # print(nonZerosInd, numNonZeros)
        nonZeroElem = np.concatenate((nonZerosInd[0].reshape(numNonZeros, 1), nonZerosInd[1].reshape(numNonZeros, 1)),
                                     axis=1)

        for step in range(self.steps):
            #print(G, G.shape)
            #print(bg, bg.shape)
            if self.calculateLossFunc:
                for element in nonZeroElem:
                    b1 = element[0]
                    b2 = element[1]

                    LossFuncCG = LossFunctionFano(u[b1][b2], b[b1][b2], A[b1][b2], G, b1, b2, self.lambda2, self.dataweightControl, self.verbose)
                    LossFunc = LossFunc + LossFuncCG
            else:
                LossFunc = (step + 1) * self.eps

            if abs(LossFunc - LossFuncGre) < self.eps or np.isnan(LossFunc):
                Lossf.append(LossFunc)
                print("already converge")
                break
            else:
                Lossf.append(LossFunc)
                LossFuncGre = LossFunc
                LossFunc = 0

                np.random.shuffle(nonZeroElem)

                batchElements = nonZeroElem[0:self.batchSize - 1, :]
                for element in batchElements:
                    b1 = element[0]
                    b2 = element[1]
                    if u[b1][b2] <= 0.01:
                        u[b1][b2] = 0.01
                    if bg[b1][0] <= 1e-09:
                        bg[b1][0] = 1e-09
                    if bg[b2][0] <= 1e-09:
                        bg[b2][0] = 1e-09
                    b[b1][b2] = np.sqrt(bg[b1][0] * bg[b2][0])

                    dG1, dG2, db1, db2 = FanoOptimize_sym(A, G, u, bg, b, b1, b2, self.lambda2, self.dataweightControl)
                    # print(dG1, dG2, db1, db2, A[b1][b2], self.lambda2)
                    G[b1, :] = G[b1, :] + alpha * dG1
                    G[b2, :] = G[b2, :] + alpha * dG2
                    bg[b1, :] = bg[b1, :] + alpha * db1
                    bg[b2, :] = bg[b2, :] + alpha * db2

                u = np.dot(G, G.T)
                b = np.sqrt(np.dot(bg, bg.transpose()))

                alpha = alpha * (1 - float(step / self.steps))
                print("number of iteration: ", step + 1, "/", self.steps, LossFuncGre)
        u = np.dot(G, G.T)
        u[u < 0] = 0

        if Lossf[-1] > Lossf[0]:
            return u, G
        elif self.dataweightControl < 2:
            self.dataweightControl += 0.5
            print("Fail to perform SMURF with this setting on the data, obj get worse. change weight as {}, learningRate as {}".format(self.dataweightControl, self.alpha))
            return self.MFFano()
        else:
            self.dataweightControl = 0
            self.alpha = 0.1 * self.alpha
            print("Fail to perform SMURF with this setting on the data, obj get worse. change weight as {}, learningRate as {}".format(self.dataweightControl, self.alpha))
            return self.MFFano()

    def MFConstCoeffiVariation(self):
        G = self.G.copy()
        A = self.A.copy()
        u = np.dot(G, G.T)

        if self.calculateIntialNoiseFactor:
            ag = geta(self.A, u)    # (Nbin, 1)
        else:
            ag = np.ones((self.bins, 1))

        a = np.sqrt(np.dot(ag, ag.transpose())) # (Nbin, Nbin)

        LossFunc = 0
        LossFuncGre = 0
        Lossf = []

        nonZerosInd = np.nonzero(np.triu(A))
        numNonZeros = nonZerosInd[0].size
        nonZeroElem = np.concatenate((nonZerosInd[0].reshape(numNonZeros, 1), nonZerosInd[1].reshape(numNonZeros, 1)),
                                     axis=1)

        for step in range(self.steps):
            if self.calculateLossFunc:
                for element in nonZeroElem:
                    b1 = element[0]
                    b2 = element[1]
                    LossFuncCG = LossFunctionConstantCoefficientVariation(u[b1][b2], a[b1][b2], A[b1][b2], G, b1, b2, self.lambda2, self.dataweightControl, self.verbose)
                    LossFunc = LossFunc + LossFuncCG
            else:
                LossFunc = (step + 1) * self.eps

            if abs(LossFunc - LossFuncGre) < self.eps:
                Lossf.append(LossFunc)
                print("already converge")
                break
            else:
                Lossf.append(LossFunc)
                LossFuncGre = LossFunc
                LossFunc = 0

                np.random.shuffle(nonZeroElem)

                batchElements = nonZeroElem[0:self.batchSize - 1, :]
                for element in batchElements:
                    b1 = element[0]
                    b2 = element[1]
                    if u[b1][b2] <= 0.01:
                        u[b1][b2] = 0.01
                    if ag[b1][0] <= 1e-09:
                        ag[b1][0] = 1e-09
                    if ag[b2][0] <= 1e-09:
                        ag[b2][0] = 1e-09
                    a[b1][b2] = np.sqrt(ag[b1][0]*ag[b2][0])

                    dG1, dG2, da1, da2 = CCVOptimize_sym(A, G, u, ag, a, b1, b2, self.lambda2, self.dataweightControl)
                    G[b1, :] = G[b1, :] + self.alpha * dG1
                    G[b2, :] = G[b2, :] + self.alpha * dG2
                    ag[b1, :] = ag[b1, :] + self.alpha * da1
                    ag[b2, :] = ag[b2, :] + self.alpha * da2

                u = np.dot(G, G.T)
                a = np.sqrt(np.dot(ag, ag.transpose()))
            print("number of iteration: ", step + 1, "/", self.steps, LossFuncGre)
        u = np.dot(G, G.T)
        u[u < 0] = 0

        return u, G

    def smurf_impute(self, initialDataFrame):
        self.A = initialDataFrame.values
        self.binNames = initialDataFrame._stat_axis.values.tolist()
        self.bins = self.A.shape[0]
        if not self.K:
            self.K = self.bins
            self.batchSize = self.K * 10
        self.genes = self.cells = self.bins

        print("Running SCEnd_sym on {} bins, with {} dimension".format(self.bins, self.K))

        self.G = initialMatrix_sym(self.A, self.K)  # (#bin, #feature)

        self._check_params()

        print("preprocessing data...")

        if self.noise_model == "CV":  # constance variance for each bin, v_b
            u, G = self.MFConstantVariance()

        elif self.noise_model == "Fano":
            u, G = self.MFFano()

        elif self.noise_model == "CCV":
            u, G = self.MFConstCoeffiVariation()

        newDataFrame = pd.DataFrame(u, index=self.binNames)

        res = {}

        res["estimate"] = newDataFrame
        res["bin latent factor matrix"] = pd.DataFrame(G, index=self.binNames, columns=None)

        self.glfm = G

        if self.estmate_only:
            return res["estimate"]
        else:
            return res

    def smurf_cell_circle(self, cells_data=None, n_neighbors=20, min_dist=0.01, major_axis=3, minor_axis=2, k=0.2):
        self.n_neighbors = n_neighbors
        self.min_dist = min_dist
        self.major_axis = major_axis
        self.minor_axis = minor_axis
        self.k = k

        if cells_data:
            data = cells_data
        else:
            if self.clfm.all():
                data = self.clfm
            else:
                raise AttributeError("Cells Data Expected")

        cell_circle_mapper = CellCircle(n_neighbors=self.n_neighbors, min_dist=self.min_dist)
        res = cell_circle_mapper.cal_cell_circle(data, a=self.major_axis, b=self.minor_axis, k=0.2)

        return res































import os
import sys
import numpy as np
import pandas as pd
from numpy import exp
from scipy import stats
import KR_norm_juicer
from sklearn.neighbors import KernelDensity
from scipy import signal, stats
import warnings
warnings.filterwarnings('ignore')

work_path = sys.argv[1]
min_ploidy = float(sys.argv[2])
max_ploidy = float(sys.argv[3])
K = sys.argv[4]
sigma = sys.argv[5]

def read_matrix(file_name):
    f = open(file_name)
    matrix = []
    chr_name = []
    bin_list = []
    count = 0
    for line in f:
        line = line.strip("\n").rstrip("\t").split("\t")
        count += 1
        if not count == 1:
            matrix.append(line[3:])
            chr_name.append(line[0])
            bin_list.append(line[0]+":"+line[1]+"-"+line[2])
        else:
            sample_list = line
    matrix = np.array(matrix)
    bin_list = np.array(bin_list)
    sample_list = np.array(sample_list[3:])
    matrix = matrix.astype(np.int)
    for i in range(len(sample_list)):
        sample_list[i] = str(sample_list[i]).split("/")[-1]
    chr_name = np.array(chr_name)
    f.close()
    return matrix, chr_name, bin_list, sample_list

def sose(scnp):
    rcnp = np.round(scnp)

    return np.sum(np.square(rcnp-scnp))

def call_CN(Y):
    Y = Y.T
    Y_nor = []
    X = np.arange(min_ploidy, max_ploidy, 0.05)
    for cell in Y:
        RCNP = cell / np.mean(cell)
        sose_list = []
        for i in X:
            sose_list.append(sose(RCNP * i))
        ploidy = X[sose_list.index(min(sose_list))]
        SCNP = RCNP * ploidy
        Y_nor.append(SCNP)
    Y_nor = np.array(Y_nor)
    Y_nor = np.round(Y_nor)
    for i in range(len(Y_nor)):
        for j in range(len(Y_nor[0])):
            if Y_nor[i][j] > 10:
                Y_nor[i][j] = 10

    return Y_nor.T

def Gaussian_kernel(data,mu,sig):
    temp = np.array(-(data - mu) ** 2 / (2 * sig ** 2), dtype="float")
    density = np.exp(temp)

    return density

def get_norm_cell(cov_data):
    cov_std = stats.variation(cov_data, axis=0)
    kde = KernelDensity(kernel="gaussian", bandwidth=0.01).fit(cov_std.reshape(-1, 1))
    dens = np.exp(kde.score_samples(np.arange(0, 1, 0.001).reshape(-1, 1)))
    peaks_pos = signal.argrelextrema(dens, np.greater)[0]
    first_peak = np.arange(0, 1, 0.001)[peaks_pos][0]
    if first_peak < 0.2:
        density_list = Gaussian_kernel(cov_std, first_peak, 0.01)
        norm_index = []
        abnorm_index = []
        for i in range(len(cov_std)):
            if density_list[i] > 1e-3:
                norm_index.append(i)
            else:
                abnorm_index.append(i)
    else:
        print("Warning: No normal cell detected! Use the reference to correct bias. Please make sure the bin size is 50 kb.")
        norm_index = []
        abnorm_index = []
        for i in range(len(cov_std)):
            abnorm_index.append(i)

    return cov_std, norm_index, abnorm_index

def Bias_norm(Y, norm_cell_index):
    Y = Y.T
    if len(norm_cell_index) > 0:
        norm_cell_Y = Y[norm_cell_index]
        bias_matrix = []
        for cell in norm_cell_Y:
            bias_list = []
            median = np.median(cell)
            for bin in cell:
                bias = bin/median
                bias_list.append(bias)
            bias_list = np.array(bias_list)
            bias_matrix.append(bias_list)
        bias_matrix = np.array(bias_matrix)
        ave_bias = bias_matrix.mean(axis=0)
        ave_bias = np.where(ave_bias==0, 1, ave_bias)
    else:
        ave_bias = np.loadtxt("bias.txt")
    
    gc_nor_Y = Y / ave_bias

    return gc_nor_Y.T

def make_adj_matrix(matrix, K, sigma):
    if K == "auto_set":
        K = 5
    else:
        K = int(K)
    if sigma == "auto_set":
        sigma = np.median(matrix)/2
    else:
        sigma = float(sigma)

    adj_matrix = np.zeros((matrix.shape[0], matrix.shape[0]))
    for i in range(matrix.shape[0]):
        for j in range(i+1, matrix.shape[0]):
            temp = (matrix[i]-matrix[j]) ** 2
            temp.sort()
            temp = temp[-K:]
            adj_matrix[i][j] = np.exp(-np.sum(temp)/sigma**2)
            adj_matrix[j][i] = np.exp(-np.sum(temp)/sigma**2)
    adj_matrix = KR_norm_juicer.KR_norm(adj_matrix)

    return adj_matrix

class DP_process:
    def __init__(self,input_matrix,K_max):
        self.matrix = input_matrix
        self.n = len(input_matrix)
        self.max_k = K_max
        self.dp_table = np.zeros((self.n, self.max_k))
        self.index_table = np.zeros((self.n, self.max_k))
        self.pre_table = np.zeros((self.n, self.n))
        for start in range(0, self.n):
            for end in range(start, self.n):
                current_matrix = self.matrix[start:end + 1, start:end + 1]
                self.pre_table[start][end] = 2 * np.sum(np.triu(current_matrix,1))
                self.pre_table[end][start] = np.sum(self.matrix[0:start, start:end + 1]) \
                                             + np.sum(self.matrix[end + 1:self.n, start:end + 1])
        self.sum = self.pre_table[0][self.n - 1]
        self.dlogd_sum = np.zeros(self.n)
        self.dlogd_sum[0] = self.pre_table[0][0] * np.log2(self.pre_table[0][0])
        for bin in range(1, self.n):
            self.dlogd_sum[bin] = self.dlogd_sum[bin - 1] + self.pre_table[bin][bin] * np.log2(self.pre_table[bin][bin]+1e-7)

    def calculate_se(self, g, V_p, V):
        if (V == 0):
            return 0
        else:
            return g / self.sum * np.log2(V_p / V)

    #record the initial state of dynamic programming (K = 0)
    def dp_init(self):
        for i in range(0, self.n):
            if i == 0:
                current_volume = self.pre_table[0][0]
            else:
                current_volume = self.pre_table[0][i] + self.pre_table[i][0]
            intra_se = (current_volume * np.log2(current_volume) - self.dlogd_sum[i])/self.sum
            inter_se = self.calculate_se(self.pre_table[i][0], self.sum, current_volume)
            self.dp_table[i][0] = inter_se + intra_se

    #dynamic process
    def dp_process(self):
        for temp_k in range(1, self.max_k):
            for temp_n in range(0, self.n):
                min_se = np.inf
                min_index = 0
                for i in range(0,temp_n):
                    se_1 =  self.dp_table[i][temp_k - 1]  # T(i,k-1)
                    if i + 1 == temp_n:
                        current_volume = self.pre_table[temp_n][temp_n]
                    else:
                        current_volume = self.pre_table[i+1][temp_n] + self.pre_table[temp_n][i+1]
                    intra_se = (current_volume * np.log2(current_volume) - (self.dlogd_sum[temp_n]-self.dlogd_sum[i])) / self.sum
                    inter_se = self.calculate_se(self.pre_table[temp_n][i + 1], self.sum, current_volume)
                    se_2 = intra_se + inter_se  # H(i+1;n)
                    temp_se = se_1 + se_2
                    if temp_se < min_se:
                        min_se = temp_se
                        min_index = i
                self.dp_table[temp_n][temp_k] = min_se
                self.index_table[temp_n][temp_k] = min_index

    #find the best K
    def find_k(self):
        k_list = self.dp_table[-1].tolist()
        self.k = k_list.index(np.min(k_list)) + 1
        self.min_entropy  = np.min(k_list)


    #find the boundaries:
    def find_boundaries(self):
        self.boundaries = []
        self.boundaries.append(int(self.n - 1))
        for i in range(1, self.k):
            self.boundaries.append(int(self.index_table[self.boundaries[i-1]][self.k-i]))
        self.boundaries.reverse()
        print(self.boundaries)
        return self.boundaries


def average_seg(boundaries, chrom_matrix):
    chrom_matrix = chrom_matrix.T
    for i in range(len(chrom_matrix)):
        for j in range(len(boundaries)):
            if j == 0:
                chrom_matrix[i][0:boundaries[j]+1] = np.median(chrom_matrix[i][0:boundaries[j]+1])
            else:
                chrom_matrix[i][boundaries[j-1]+1:boundaries[j]+1] = np.median(chrom_matrix[i][boundaries[j-1]+1:boundaries[j]+1])

    return chrom_matrix.T


def save_matrix(matrix, bin_list, sample_list, var, cnv_name, meta_name):
    ploidy = np.average(matrix, axis=1)
    ploidy = np.round(ploidy, 1)
    df = pd.DataFrame(matrix, columns=bin_list, index=sample_list)
    df.to_csv(cnv_name)
    df_2 = pd.DataFrame(np.vstack((sample_list, ploidy, var)).T, columns=["cell_id", "c_ploidy", "c_variance"])
    df_2.to_csv(meta_name, index=False)

def main():

    Y, chr_name, bin_list, sample_list = read_matrix(os.path.join(work_path, "genome_cov.bed"))
    print("Begin")
    var, norm_cell_index, abnorm_cell_index = get_norm_cell(Y)
    cov_matrix = Bias_norm(Y, norm_cell_index)
    chosen_chr = ["chr1", "chr2", "chr3", "chr4", "chr5", "chr6", "chr7", "chr8", "chr9", "chr10", "chr11", "chr12",
                  "chr13", "chr14", "chr15", "chr16", "chr17", "chr18", "chr19", "chr20", "chr21", "chr22"]
    for chrom in chosen_chr:
        print("process %s..."%chrom)
        indexes = np.where(chr_name == chrom)
        chrom_matrix = cov_matrix[indexes]
        matrix = chrom_matrix[:, abnorm_cell_index]
        adj_matrix = make_adj_matrix(matrix, K, sigma)
        dp_process = DP_process(adj_matrix, 20)
        dp_process.dp_init()
        dp_process.dp_process()
        dp_process.find_k()
        boundaries = dp_process.find_boundaries()
        new_chrom = average_seg(boundaries, chrom_matrix)
        if chrom == "chr1":
            new_cov_matrix = new_chrom
        else:
            new_cov_matrix = np.vstack((new_cov_matrix, new_chrom))
    cn_matrix = call_CN(new_cov_matrix)
    save_matrix(cn_matrix.T, bin_list, sample_list, var, os.path.join(work_path, "cnv_matrix.csv"), os.path.join(work_path, "cnv_meta.csv"))


if __name__ == "__main__":
    main()




import numpy as np
from math import log2
from scipy.cluster import hierarchy
from sklearn.cluster import AgglomerativeClustering
from sklearn.neighbors import kneighbors_graph
from scipy import sparse
import heapq
import itertools
import networkx as nx
from itertools import chain
import pandas as pd
from queue import Queue
import kmeans1d
import scipy.linalg as la
from scipy.sparse import csgraph
# import umap
import time

# import seat_wrapper

from . import se
# from .KR_norm_juicer import KR_norm

class DP_process:
    def __init__(self,input_matrix,K_max):
        self.matrix = input_matrix
        self.n = len(input_matrix)
        self.max_k = K_max
        self.dp_table = np.zeros((self.n, self.max_k))
        self.index_table = np.zeros((self.n, self.max_k))
        self.pre_table = np.zeros((self.n, self.n))
        for start in range(0, self.n):
            for end in range(start, self.n):
                current_matrix = self.matrix[start:end + 1, start:end + 1]
                self.pre_table[start][end] = 2 * np.sum(np.triu(current_matrix,1))
                self.pre_table[end][start] = np.sum(self.matrix[0:start, start:end + 1]) \
                                             + np.sum(self.matrix[end + 1:self.n, start:end + 1])
        self.sum = self.pre_table[0][self.n - 1]
        self.dlogd_sum = np.zeros(self.n)
        self.dlogd_sum[0] = self.pre_table[0][0] * np.log2(self.pre_table[0][0])
        for bin in range(1, self.n):
            self.dlogd_sum[bin] = self.dlogd_sum[bin - 1] + self.pre_table[bin][bin] * np.log2(self.pre_table[bin][bin]+1e-7)

    def calculate_se(self, g, V_p, V):
        if (V == 0):
            return 0
        else:
            return g / self.sum * np.log2(V_p / V)

    #record the initial state of dynamic programming (K = 0)
    def dp_init(self):
        for i in range(0, self.n):
            if i == 0:
                current_volume = self.pre_table[0][0]
            else:
                current_volume = self.pre_table[0][i] + self.pre_table[i][0]
            intra_se = (current_volume * np.log2(current_volume) - self.dlogd_sum[i])/self.sum
            inter_se = self.calculate_se(self.pre_table[i][0], self.sum, current_volume)
            self.dp_table[i][0] = inter_se + intra_se

    #dynamic process
    def dp_process(self):
        for temp_k in range(1, self.max_k):
            for temp_n in range(0, self.n):
                min_se = np.inf
                min_index = 0
                for i in range(0,temp_n):
                    se_1 =  self.dp_table[i][temp_k - 1]  # T(i,k-1)
                    if i + 1 == temp_n:
                        current_volume = self.pre_table[temp_n][temp_n]
                    else:
                        current_volume = self.pre_table[i+1][temp_n] + self.pre_table[temp_n][i+1]
                    intra_se = (current_volume * np.log2(current_volume) - (self.dlogd_sum[temp_n]-self.dlogd_sum[i])) / self.sum
                    inter_se = self.calculate_se(self.pre_table[temp_n][i + 1], self.sum, current_volume)
                    se_2 = intra_se + inter_se  # H(i+1;n)
                    temp_se = se_1 + se_2
                    if temp_se < min_se:
                        min_se = temp_se
                        min_index = i
                self.dp_table[temp_n][temp_k] = min_se
                self.index_table[temp_n][temp_k] = min_index

    #find the best K
    def find_k(self):
        k_list = self.dp_table[-1].tolist()
        self.k = k_list.index(np.min(k_list)) + 1
        self.min_entropy  = np.min(k_list)


    #find the boundaries:
    def find_boundaries(self):
        self.boundaries = []
        self.boundaries.append(int(self.n - 1))
        for i in range(1, self.k):
            self.boundaries.append(int(self.index_table[self.boundaries[i-1]][self.k-i]))
        self.boundaries.reverse()
        print(self.boundaries)
        return self.boundaries


class Node():
    def __init__(self, graph_stats, node_id, parent=None, leaf=False,
                 is_singleton=True,
                 is_leaf=True):
        self.id = node_id
        self.parent = parent
        self.leaf = leaf
        self.is_singleton = is_singleton
        self.is_leaf = is_leaf
        self.children = []
        self.left = 0
        self.right = 0
        self.g = 0.
        self.g_log_V = 0.
        self.V = 0.
        self.log_V = 0.
        self.V_log_V = 0.
        self.s = 0.
        self.se = 0.
        self.vs = []
        self.sum_b_log_b = 0.
        self.height = 0
        self.dist = 1
        self.graph_stats = graph_stats

    def __repr__(self):
        return self.toString()
        # return "Node()"

    def __str__(self):
        return self.toString()

    def toString(self):
        return "(idx:%d, %sdist:%d, [%d, %d], len(children)=%d, len(vs):%d)" % (self.id,
                                                                                'leaf, ' if self.leaf else '',
                                                                                self.dist, self.left, self.right,
                                                                                len(self.children),
                                                                                len(self.vs))

    def get_max_height(self):
        max_h = 0
        if not self.leaf:
            for child in self.children:
                if isinstance(child, Node):
                    h_tmp = child.get_max_height()
                    max_h = max(max_h, h_tmp)
        else:
            return self.height
        return max_h

    def reset(self, parent=None):
        self.init(parent)

    def init(self, parent=None, setBins=True):
        if setBins:
            self.setBins()
        self.setV()
        self.setS()
        self.setG()
        if parent:
            self.setSE(parent)
        else:
            # root
            self.se = 0
        for c in self.children:
            if (isinstance(c, Node)):
                c.setSE(self)

    def setBins(self):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        vs = []
        if self.leaf:
            vs += self.children
            self.sum_b_log_b = np.sum(d_log_d[self.vs])
        else:
            for c in self.children:
                if isinstance(c, Node):
                    vs += c.setBins()
                elif isinstance(c, int):
                    vs.append(c)
                else:
                    raise TypeError('child can only be int or Node')
        self.vs = vs
        self.left = self.vs[0]
        self.right = self.vs[-1]
        return vs

    def setS(self):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        self.s = se.get_s(M, sparse_m, self.vs)

    def setG(self):
        self.g = self.V - self.s
        self.g_log_V = self.g * self.log_V
        # self.g = getG(self.vs)

    def setV(self):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        self.V = se.get_v(M, sparse_m, self.vs)
        self.log_V = log2(self.V)
        self.V_log_V = self.V * self.log_V

    def setSE(self, parent):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        self.se = se.get_se(vG, self.g, self.V, parent.V)

    def increase_height(self, increment=1):
        self.height += increment
        for c in self.children:
            if isinstance(c, Node):
                c.increase_height(increment)

    def merge(self, node_id, node1, node2, is_leaf=False):
        if (node1.parent != node2.parent):
            print("self=", self, "\nnode1=", node1, node1.parent, "\nnode2=", node2, node2.parent)
            raise ValueError("parents are not the same")
        node = Node(self.graph_stats, node_id, parent=self.id)
        node.leaf = is_leaf
        node.height = node1.height
        node1.parent = node.id
        node1.increase_height(1)
        node2.parent = node.id
        node2.increase_height(1)
        if is_leaf:
            node.children = node1.children + node2.children
        else:
            node.addChild(node1)
            node.addChild(node2)
            node.dist = max(node1.dist, node2.dist) + 1
        # print("before se=", node.se)
        node.reset(self)
        # print("after se=", node.se)
        self.addChild(node)
        if not self.delChild(node1):
            raise("fail to delete child", node1)
        if not self.delChild(node2):
            raise("fail to delete child", node1)
        # print("combine:\nself=", self, "\nnew node=", node, "\nnode1=", node1, "\nnode2=", node2)
        # print("node1.height=%d, node2.height=%d" % (node1.height, node2.height))
        return node

    def addChild(self, node):
        self.children.append(node)

    def delChild(self, node):
        idx = -1
        for i, c in enumerate(self.children):
            if c.id == node.id:
                idx = i
                break
        if idx >= 0:
            del self.children[idx]
            return True
        return False

    def printChildren(self):
        if all(isinstance(c, int) for c in self.children):
            print("\t".join(self.children))
        else:
            print("\t".join([c.id for c in self.children]))

    def verbose(self, k):
        print('-'*(k+1) + str(self))
        if len(self.children) > 0:
            for c in self.children:
                if isinstance(c, Node):
                    c.verbose(k+2)
                elif isinstance(c, int):
                    print("*"*(k+1+2) + str(c))
                else:
                    raise TypeError('child cannot be int or Node')

    def getNodes(self):
        nodes = []
        if not self.leaf:
            for c in self.children:
                nodes.append(c.vs)
                if isinstance(c, Node):
                    nodes.extend(c.getNodes())
                elif isinstance(c, int):
                    nodes.extend([c])
                else:
                    raise TypeError('child can only be int or Node')
        return nodes

    def getTreeSE(self):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        se = self.se
        # print(self, "g=", self.g, "v=", self.V, "se=", self.se, "vG=", vG)
        if self.leaf:
            se += (self.V_log_V - np.sum(d_log_d[self.children]))/vG
        else:
            for c in self.children:
                if isinstance(c, Node):
                    se += c.getTreeSE()
                elif isinstance(c, int):
                    print("getSEforBin(c, self.V)=", se.get_se_for_vertex(M, d, vG, c, self.V))
                    # se += getSEforBin(c, self.V)
                    se += (d[c]*self.log_V - d_log_d[c])/vG
                else:
                    raise TypeError('child can only be int or Node')
        return se


class pySETree():

    def __init__(self, aff_m, min_k=2, max_k=10,
                 max_g_ratio=0,
                 objective='structure_entropy',
                 strategy='top_down'):
        self.max_g_ratio = max_g_ratio
        self.strategy = strategy
        self.objective = objective
        self.min_k = min_k
        self.max_k = max_k

        self.vertex_num = aff_m.shape[0]
        if self.max_k > self.vertex_num:
            self.max_k = self.vertex_num - 1

        self.ks = range(self.min_k, self.max_k+1)

        if strategy == 'top_down':
            self.node_id = 2*self.vertex_num - 3
        else:
            self.node_id = -2
        self.node_list = {}

        self.affinity_m = aff_m
        self.graph_stats = self.graph_stats_init(aff_m)

    def graph_stats_init(self, aff_m):
        M = aff_m
        np.fill_diagonal(M, 0)
        d = np.sum(M, 1) - M.diagonal()
        if np.any(d == 0):
            M += 1e-3
            np.fill_diagonal(M, 0)
            d = np.sum(M, 1) - M.diagonal()
        # print("d=", d)
        log_d = np.log2(d)
        d_log_d = np.multiply(d, log_d)
        # print("sum(d)=", sum(d))

        sparce_m = sparse.csr_matrix(aff_m)
        # m = np.sum(np.triu(M, 1))
        m = sparce_m.sum() / 2
        # print("m=", m)
        # vG = np.sum(M)
        vG = sparce_m.sum()
        # print("vG=", vG)
        log_vG = log2(vG)

        graph_stats = M, m, d, log_d, d_log_d, vG, log_vG, sparce_m
        return graph_stats

    def get_node_id(self, increment=True):
        if increment:
            self.node_id += 1
        else:
            self.node_id -= 1
        return self.node_id

    def build_tree(self):
        M, m, d, log_d, d_log_d, vG, log_vG, sparce_m = self.graph_stats
        root = Node(self.graph_stats, self.get_node_id())
        self.node_list[root.id] = root
        root.V = 2.*m

        print(self.strategy)
        if self.strategy == 'bottom_up':
            Z = self.bottom_up(root)
        else:
            Z = self.top_down(root)

        return Z

    def bottom_up(self, root):
        for i in range(self.vertex_num):
            node = Node(self.graph_stats, self.get_node_id(), parent=root.id, leaf=True)
            self.node_list[node.id] = node
            node.height = 1
            node.children = [i]
            node.left = node.right = i
            node.init(root)
            root.addChild(node)
        root.reset()
        # root.verbose(0)
        Z = self.linkage(root, tree_type='multinary', by='heap')
        self.Z = Z[:, :4]
        return Z

    def top_down(self, root):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats

        N = self.vertex_num
        root.vs = np.array(range(N))
        Z = np.zeros((N - 1, 5))

        self.leafs = []
        nodes_to_divide = Queue(maxsize=N)
        nodes_to_divide.put(root)
        while not nodes_to_divide.empty():
            node = nodes_to_divide.get()
            self.dividing_tree(node, nodes_to_divide, Z)

        # root.verbose(0)
        height = max(Z[:, 2])
        Z[np.argwhere(Z[:, 2] == -1), 2] = height + 1
        Z[:, 2] = height + 2 - Z[:, 2]
        return Z

    def _get_dividing_delta(self, node, children):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        if self.objective == 'modularity':
            delta = (children[0].s/vG - np.power(children[0].V/vG, 2) + children[1].s/vG - np.power(children[1].V/vG, 2))
        else:
            # delta = (children[0].se + children[0].V_log_V + children[1].se + children[1].V_log_V) - node.se - node.V_log_V
            # delta = node.se - (children[0].se + children[1].se)
            # print(children[0].se, children[1].se, node.se)
            if self.max_g_ratio:
                delta = 0
                for child in children:
                    g = max(child.g - self.vertex_num*(self.vertex_num-len(child.vs))*self.max_g_ratio, 0)
                    delta += se.get_se(vG, g, child.V, node.V)
            else:
                delta = children[0].se + children[1].se
        return delta

    def dividing_tree(self, node, nodes_to_divide, Z):

        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats

        if len(node.vs) > 2:
            A = M[np.ix_(node.vs, node.vs)]
            '''
            A = A + 0.001
            np.fill_diagonal(A, 0)
            A_KR = KR_norm(A)
            '''
            L = csgraph.laplacian(A, normed=True)
            eig_values, eig_vectors = la.eigh(L)

            unique_eig_values = np.sort(list(set(eig_values.real)))
            if len(unique_eig_values) == 1:
                mid = int(len(node.vs)/2)
                node1_vs = range(mid)
                node2_vs = range(mid, len(node.vs))
            else:
                fiedler_pos = np.where(eig_values.real == unique_eig_values[1])[0][0]
                fiedler_vector = np.transpose(eig_vectors)[fiedler_pos]
                clusters, centroids = kmeans1d.cluster(fiedler_vector, 2)
                node1_vs = np.argwhere(np.array(clusters) == 0).T[0]
                node2_vs = np.argwhere(np.array(clusters) == 1).T[0]
            '''
            print('eig_values', eig_values)
            print('unique eigen values', unique_eig_values)
            print('eigen vector', eig_vectors)
            print(list(set(eig_values.real)))
            print(np.sort(list(set(eig_values.real))))
            '''

            '''
            cutoff = fiedler_vector.mean()
            cutoff = 0
            print(fiedler_pos, "Fiedler vector: " + str(fiedler_vector.real), cutoff)
            node1_vs = np.argwhere(fiedler_vector > cutoff).T[0]
            node2_vs = np.argwhere(fiedler_vector <= cutoff).T[0]
            '''
        else:
            node1_vs = [0]
            node2_vs = [1]

        children_vs = [node1_vs, node2_vs]

        children = []
        for child_vs in children_vs:
            child_vs = node.vs[child_vs]
            if len(child_vs) == 1:
                child_id = child_vs[0]
            else:
                child_id = self.get_node_id(increment=False)
            child = Node(self.graph_stats, child_id, parent=node.id)
            child.vs = child_vs
            child.init(parent=node, setBins=False)
            self.node_list[child.id] = child
            node.addChild(child)
            children.append(child)
            child.height = node.height + 1

            if len(child.vs) > 1:
                nodes_to_divide.put(child)
            else:
                child.leaf = True

        delta = self._get_dividing_delta(node, children)
        if delta > 0 and self.vertex_num != len(node.vs):  # not split
            node.leaf = True
        if (node.parent and self.node_list[node.parent].leaf) or len(node.vs) == 2:
            node.leaf = True

        if node.leaf:
            for child in children:
                child.height = -1
                child.leaf = True
        # print(node.id, node.leaf, children[0].id, children[0].leaf, children[1].id, children[1].leaf, node.se, delta)

        for n in [node] + children:
            if n.leaf and not self.node_list[n.parent].leaf:  # and not self.node_list[node.parent].leaf:
                self.leafs.append(n)

        Z[node.id-(self.vertex_num)] = [children[0].id, children[1].id, children[0].height, len(node.vs), node.id]

    def get_max_delta_from_table(self, m, row_ids, col_ids):
        m = m[np.ix_(row_ids, col_ids)]
        max_i = np.argmax(m)
        row_i = int(max_i/m.shape[1])
        col_i = max_i % m.shape[1]
        max_delta = m[row_i, col_i]
        max_n1 = self.node_list[row_ids[row_i]]
        max_n2 = self.node_list[col_ids[col_i]]
        return max_n1, max_n2, max_delta

    def get_max_delta_from_heap(self, heap, row_ids, col_ids):
        while heap:
            max_delta, max_n1, max_n2 = heapq.heappop(heap)
            if max_n1 not in row_ids or max_n2 not in col_ids:
                continue
            max_n1 = self.node_list[max_n1]
            max_n2 = self.node_list[max_n2]
            max_delta = -max_delta
            return max_n1, max_n2, max_delta
        else:
            return None, None, None

    def linkage(self, root, tree_type='multinary', by='heap'):
        print('linkage', tree_type, by)
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        G = nx.from_numpy_matrix(M)

        N = self.vertex_num
        Z = np.zeros((N - 1, 5))
        delta_m = np.zeros((2*N - 1, 2*N - 1))
        delta_m.fill(-10000)
        leafs = {n: 1 for n in range(N)}
        singletons = {n: 1 for n in range(N)}

        heap = []
        heapq.heapify(heap)

        i = 0
        for n1, n2 in zip(*np.triu(M, 1).nonzero()):  # O(kn)
            node1 = self.node_list[n1]
            node2 = self.node_list[n2]
            if self.objective == 'structure_entropy':
                delta = se.get_delta_se_plus(M, sparse_m, vG, d, root, node1, node2)
            else:  # network modularity
                delta = se.get_delta_nm(M, sparse_m, vG, root, node1, node2)
            delta_m[n1, n2] = delta
            heapq.heappush(heap, (-delta, n1, n2))
            i += 1
        print('i', i, '(linkage - initial non zero pair)')

        z_i = 0

        count = 0
        for only_positive_delta in [True, False]:
            if self.objective == 'modularity':
                only_positive_delta = False
            while singletons:
                if tree_type == 'binary' and count > 0:
                    break
                print('merge phase', self.objective, only_positive_delta, count)
                z_i, merge_phase_i = self._merge_phase(G, root, singletons, leafs, Z, z_i, delta_m, heap, count+1,
                                                       only_positive_delta=only_positive_delta, by=by)
                i += merge_phase_i
                print('i', i, 'merge phase')
                count += 1
                if singletons:
                    break
                singletons = {}
                for l in leafs:
                    self.node_list[l].is_singleton = True
                    singletons[l] = 1
                    i += 1
                print('i', i, 'merge phase update leafs states')
                if not G.edges(data=True):
                    break

        self.leafs = [self.node_list[l] for l in leafs]
        while z_i < self.vertex_num - 1:
            only_positive_delta = False
            print('binary merge', self.objective, only_positive_delta, count)
            z_i, binary_merge_i = self._binary_merge(G, root, leafs, Z, z_i, delta_m,
                                                     only_positive_delta=only_positive_delta,
                                                     by=by)
            i += binary_merge_i
            print('i', i, 'binary_merge')
            count += 1

        print('N', N, 'i', i)
        return Z

    def _merge_phase(self, G, root, singletons, leafs, Z, z_i, delta_m, heap, dist,
                     only_positive_delta=True, is_leaf=True, by='heap'):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        i = 0
        print('singletons len', len(singletons))
        while singletons:
            # print('i', i)
            if not G.edges(data=True):
                return z_i, i

            if by == 'table':
                max_n1, max_n2, max_delta = self.get_max_delta_from_table(delta_m, singletons, leafs)
            else:
                max_n1, max_n2, max_delta = self.get_max_delta_from_heap(heap, singletons, leafs)

            if max_n1 is None:
                return z_i, i

            new_node = root.merge(self.get_node_id(), max_n1, max_n2, is_leaf=is_leaf)
            self.node_list[new_node.id] = new_node

            # Z[z_i] = [max_n1.id, max_n2.id, new_node.dist, len(max_n1.vs) + len(max_n2.vs), new_node.id]
            Z[z_i] = [max_n1.id, max_n2.id, dist, len(max_n1.vs) + len(max_n2.vs), new_node.id]

            # update
            del singletons[max_n1.id]
            max_n1.is_singleton = False
            if max_n2.is_singleton:
                del singletons[max_n2.id]
                max_n2.is_singleton = False
            if max_n1.is_leaf:
                del leafs[max_n1.id]
                max_n1.is_leaf = False
            del leafs[max_n2.id]
            max_n2.is_leaf = False

            new_node.is_leaf = True
            new_node.is_singleton = False

            G.add_node(new_node.id)
            # for x in singletons:
            for x in set(chain(G.neighbors(max_n1.id), G.neighbors(max_n2.id))):
                i += 1
                node = self.node_list[x]
                if self.objective == 'structure_entropy':
                    delta = se.get_delta_se_plus(M, sparse_m, vG, d, root, node, new_node)
                    # delta_se = se.get_delta_se(M, sparse_m, vG, root, node, new_node)
                else:
                    delta = se.get_delta_nm(M, sparse_m, vG, root, node, new_node)
                if only_positive_delta and delta < 0:
                    continue
                delta_m[x, new_node.id] = delta
                heapq.heappush(heap, (-delta, x, new_node.id))
                G.add_edge(x, new_node.id, weight=1)
            # print('i', i, 'update')

            leafs[new_node.id] = 1
            G.remove_node(max_n1.id)
            G.remove_node(max_n2.id)

            z_i += 1

        return z_i, i

    def _binary_merge(self, G, root, leafs, Z, z_i, delta_m, only_positive_delta=True, by='heap'):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        heap = []
        heapq.heapify(heap)
        i = 0
        ns = [(n1, n2) for n1, n2, _ in G.edges(data=True)]
        # ns = []  # not solving HC problem
        # print(leafs)
        # print(ns)
        if not ns:
            ns = itertools.combinations(leafs, 2)
        for n1, n2 in ns:
            node1, node2 = self.node_list[n1], self.node_list[n2]
            if self.objective == 'structure_entropy':
                # delta = se.get_delta_se_plus(M, sparse_m, vG, d, root, node1, node2)  # fully stairstep shape
                delta = se.get_delta_se(M, sparse_m, vG, root, node1, node2)
            else:
                delta = se.get_delta_nm(M, sparse_m, vG, root, node1, node2)
            delta_m[n1, n2] = delta
            if only_positive_delta and delta < 0:
                continue
            heapq.heappush(heap, (-delta, n1, n2))
            i += 1

        while z_i < self.vertex_num - 1:
            if by == 'table':
                max_n1, max_n2, max_delta = self.get_max_delta_from_table(delta_m, leafs, leafs)
            else:
                max_n1, max_n2, max_delta = self.get_max_delta_from_heap(heap, leafs, leafs)

            if max_n1 is None:
                return z_i, i

            new_node = root.merge(self.get_node_id(), max_n1, max_n2, is_leaf=False)
            self.node_list[new_node.id] = new_node
            # print(max_n1.id, max_n2.id, new_node.id, max_delta)

            Z[z_i] = [max_n1.id, max_n2.id, new_node.dist, len(max_n1.vs) + len(max_n2.vs), new_node.id]

            # update
            del leafs[max_n1.id]
            del leafs[max_n2.id]
            G.add_node(new_node.id)  # O(k)
            xs = set(chain(G.neighbors(max_n1.id), G.neighbors(max_n2.id)))
            # xs = []  # not solving HC problem
            if not xs:
                xs = leafs
            for x in xs:
                node = self.node_list[x]
                if self.objective == 'structure_entropy':
                    delta = se.get_delta_se(M, sparse_m, vG, root, node, new_node)
                else:
                    delta = se.get_delta_nm(M, sparse_m, vG, root, node, new_node)
                if only_positive_delta and delta < 0:
                    continue
                delta_m[x, new_node.id] = delta
                heapq.heappush(heap, (-delta, x, new_node.id))
                i += 1
            G.remove_node(max_n1.id)
            G.remove_node(max_n2.id)
            leafs[new_node.id] = 1

            z_i += 1

        # print(delta_m[262, 445], delta_m[441, 445])
        # print(delta_m[238, 262], delta_m[262, 439], delta_m[238, 439])

        return z_i, i

    def cut_tree(self, Z, n_clusters):
        # update node distance
        if self.strategy == 'bottom_up':
            root = self.node_list[self.node_id]
        else:
            root = self.node_list[2*self.vertex_num - 2]
        self.root = root
        # self._update_dist_to_level(root, Z)
        se_scores, ks_clusters, optimal_k = self._cut_tree_dp(root)
        tmp = pd.DataFrame(np.matrix(se_scores), columns=self.ks).T
        tmp['K'] = tmp.index
        tmp.columns = ['SE Score', 'K']
        self.se_scores = tmp
        delta_se_scores = se_scores[1:] - se_scores[:-1]
        tmp = pd.DataFrame(np.matrix(delta_se_scores), columns=self.ks[1:]).T
        tmp['K'] = tmp.index
        tmp.columns = ['Delta SE Score', 'K']
        self.delta_se_scores = tmp
        self.ks_clusters = pd.DataFrame(np.matrix(ks_clusters).T, columns=['K={}'.format(k) for k in self.ks])
        Z_clusters = hierarchy.cut_tree(Z[:, :4], n_clusters=n_clusters)
        self.Z_clusters = pd.DataFrame(np.matrix(Z_clusters), columns=['K={}'.format(k) for k in n_clusters])

        self.optimal_k = optimal_k
        self.optimal_clusters = self.ks_clusters['K={}'.format(self.optimal_k)].tolist()
        return

    def _cut_tree_dp(self, root):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        nodes = range(self.vertex_num*2 - 1)
        # print(nodes)  # the nodes ordered from child to parent

        # print('max_k', max_k)
        cost_m = np.zeros((len(nodes), self.max_k+1))
        # cost_m.fill(100)
        cutoff_m = np.zeros((len(nodes), self.max_k+1))
        cutoff_m.fill(-1)
        np.set_printoptions(suppress=True)

        '''
        for k in self.ks:
            self._cut_tree_dp_recursive(root, root, k, vG, cost_m, cutoff_m, nodes)
        '''
        self._dp_compute_cost(cost_m, cutoff_m, nodes)
        # print(cost_m[-1, :])
        print(cost_m)
        print(cost_m.shape)
        # print(cutoff_m)
        ks_clusters = []
        for k in self.ks:
            if k == 1:
                ks_clusters.append([0]*self.vertex_num)
                continue
            clusters = []
            # print('k', k, '----')
            self._trace_back(root, cost_m, cutoff_m, nodes, clusters, k)
            # print(clusters)
            clusters = [(v, i) for i, c in enumerate(clusters) for v in self.node_list[c].vs]
            clusters = sorted(clusters)
            clusters = [c for v, c in clusters]
            # print(len(clusters), self.vertex_num)
            if len(clusters) != self.vertex_num:  # happens in bottom up node if k larger than number of leafs
                ks_clusters.append([0]*self.vertex_num)
                continue

            ks_clusters.append(clusters)

        optimal_k = self.max_k - np.argmin(cost_m[-1, 2:][::-1])
        return cost_m[-1, 1:], ks_clusters, optimal_k

    def _cut_tree_dp_ms_backup(self, root):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        nodes = range(self.vertex_num*2 - 1)
        # print(nodes)  # the nodes ordered from child to parent

        # print('max_k', max_k)
        # cost_m = np.zeros((len(nodes), self.max_k+1))
        cost_m = np.zeros((len(nodes), 5))
        # cost_m.fill(100)
        cutoff_m = np.zeros((len(nodes), 5))
        cutoff_m.fill(-1)
        np.set_printoptions(suppress=True)

        '''
        for k in self.ks:
            self._cut_tree_dp_recursive(root, root, k, vG, cost_m, cutoff_m, nodes)
        '''
        self._dp_compute_cost(cost_m, cutoff_m, nodes)
        # print(cost_m[-1, :])
        print(cost_m)
        print(cost_m.shape)
        # print(cutoff_m)
        ks_clusters = []
        for k in self.ks:
            if k == 1:
                ks_clusters.append([0]*self.vertex_num)
                continue
            clusters = []
            # print('k', k, '----')
            self._trace_back(root, cost_m, cutoff_m, nodes, clusters, k)
            # print(clusters)
            clusters = [(v, i) for i, c in enumerate(clusters) for v in self.node_list[c].vs]
            clusters = sorted(clusters)
            clusters = [c for v, c in clusters]
            # print(len(clusters), self.vertex_num)
            if len(clusters) != self.vertex_num:  # happens in bottom up node if k larger than number of leafs
                ks_clusters.append([0]*self.vertex_num)
                continue

            ks_clusters.append(clusters)

        optimal_k = self.max_k - np.argmin(cost_m[-1, 2:][::-1])
        return cost_m[-1, 1:], ks_clusters, optimal_k

    def _trace_back(self, node, cost_m, cutoff_m, nodes, clusters, k_hat):
        if type(node) == int:
            return
        if len(node.vs) == 1:
            return
        i = node.id

        if k_hat == 1:
            return

        k_prime = int(cutoff_m[i, k_hat])
        # print(cutoff_m.shape, 'n', i, 'k_hat', k_hat, 'k_prime', k_prime)
        # print(node)
        left_node = node.children[0]
        right_node = node.children[-1]

        if k_prime > 1:
            self._trace_back(left_node, cost_m, cutoff_m, nodes, clusters, k_prime)
        else:
            if type(left_node) == int:
                clusters.append(left_node)
            else:
                clusters.append(left_node.id)
        if k_prime < k_hat-1:
            self._trace_back(right_node, cost_m, cutoff_m, nodes, clusters, k_hat-k_prime)
        else:
            if type(right_node) == int:
                clusters.append(right_node)
            else:
                clusters.append(right_node.id)

    def _dp_compute_cost(self, cost_m, cutoff_m, nodes):

        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats

        leaves = hierarchy.leaves_list(self.Z)
        order = [(l, i) for i, l in enumerate(leaves)]
        order.sort()
        order = [ l for l, i in order]
        M = M[leaves, :]
        M = M[:, leaves]
        dp_process = DP_process(M+1, 10)
        dp_process.dp_init()
        dp_process.dp_process()
        dp_process.find_k()
        boundaries = dp_process.find_boundaries()

        exit()
        for n in nodes:
            node = self.node_list[n]
            if type(node) == int:
                cost_m[node, :] = 1000
                continue

            for k in self.ks:
                if (self.strategy == 'bottom_up' and k != 1 and node.leaf) or len(node.vs) < k:
                    cost_m[node.id, k] = 1000
                    # print('n', node.id, 'k', k, 1000)
                    continue

                # print(se.get_se(vG, node.g, node.V, parent.V) == node.se)
                if self.objective == 'structure_entropy':
                    # cost = node.g/vG*log2(vG/node.V) - node.se
                    if self.max_g_ratio:
                        g = max(node.g - self.vertex_num*(self.vertex_num-len(node.vs))*self.max_g_ratio, 0)
                        parent_V = vG if not node.parent else self.node_list[node.parent].V
                        node_cost = se.get_se(vG, g, node.V, parent_V)
                    else:
                        node_cost = node.se
                else:
                    node_cost = -(node.s/vG - np.power(node.V/vG, 2))
                if k == 1:
                    cost_m[node.id, k] = node_cost + node.V/vG*(node.V*node.log_V - node.sum_b_log_b)
                    cost_m[node.id, k] = se.get_se(vG, node.g, node.V, vG) \
                        + se.get_se(vG, node.g, node.V, node.V) \
                        + node.V*node.log_V - node.sum_b_log_b
                    # cost_m[node.id, k] = node_cost
                    # print('n', node.id, 'k', k, cost)
                    continue

                if type(node.children[0]) == int:
                    l_id = node.children[0]
                else:
                    l_id = node.children[0].id
                if type(node.children[1]) == int:
                    r_id = node.children[1]
                else:
                    r_id = node.children[1].id
                min_i = None
                min_cost = 100000
                for i in range(1, k):
                    cost = cost_m[l_id, i] + cost_m[r_id, k-i] # + node_cost
                    if cost < min_cost:
                        min_cost = cost
                        min_i = i

                # print('k', k, 'min_cost', min_cost, node, l_id, r_id, min_i)
                cost_m[node.id, k] = min_cost
                cutoff_m[node.id, k] = min_i
                # print('n', node.id, 'k', k, min_cost)

    def _cut_tree_dp_recursive(self, parent, node, k, vG, cost_m, cutoff_m, nodes):
        M, m, d, log_d, d_log_d, vG, log_vG, sparse_m = self.graph_stats
        if type(node) == int:
            return 1000
        if len(node.vs) < k:
            return 1000
        if k == 1:
            # print(se.get_se(vG, node.g, node.V, parent.V) == node.se)
            if self.objective == 'structure_entropy':
                cost = node.se
                # cost = node.g/vG*log2(vG/node.V) - node.se
            else:
                cost = -(node.s/vG - np.power(node.V/vG, 2))
            '''
            if not node.leaf:
                cost_m[k-1, nodes.index(node.id)]
            '''
            return cost

        min_cost = 100000000
        min_i = None
        for i in range(1, k):
            '''
            if type(node.children[0]) == int:
                l_id = node.children[0]
            else:
                l_id = node.children[0].id
            if np.isnan(cost_m[l_id, i]):
                l_cost = self._cut_tree_dp_recursive(node, node.children[0], i, vG, cost_m, cutoff_m, nodes)
            else:
                l_cost = cost_m[l_id, i]
            '''
            l_cost = self._cut_tree_dp_recursive(node, node.children[0], i, vG, cost_m, cutoff_m, nodes)
            r_cost = self._cut_tree_dp_recursive(node, node.children[1], k-i, vG, cost_m, cutoff_m, nodes)
            cost = l_cost + r_cost + node.se
            # print(k, i, cost)
            if cost < min_cost:
                min_cost = cost
                min_i = i
        # print('k', k, 'min_cost', min_cost, node, node.children[0], node.children[1], min_i, k)
        cost_m[node.id, k] = min_cost
        cutoff_m[node.id, k] = min_i
        return min_cost

    def _update_dist_to_level(self, parent, Z):
        for c in parent.children:
            c.reset(parent)
            if c.leaf:
                c.reset(parent)
                continue
            c.dist = parent.dist - 1
            if c.id > self.vertex_num:
                Z[c.id-self.vertex_num, 2] = c.dist
            self._update_dist_to_level(c, Z)

    def to_newick(self):
        return '({});'.format(self._to_newick_aux(self.root, is_root=True))

    def _to_newick_aux(self, node, is_root=False):
        if type(node) == int or type(node) == np.int32:
            return 'n{}:{}'.format(node, 1)
        if len(node.vs) == 1:
            return 'n{}:{}'.format(node.id, 1)

        if node.leaf:
            if self.strategy == 'bottom_up':
                res = self._to_newick_leaf_bottom_up(node)
            else:
                res = self._to_newick_leaf_top_down(node)
        else:
            res = ','.join([self._to_newick_aux(c) for c in node.children])
        if is_root:
            res = '({})n{}:{}'.format(res, node.id, 1)
        else:
            res = '({})n{}:{}'.format(res, node.id, 1)
        return res

    def _to_newick_leaf_bottom_up(self, node):
        if type(node) == int or type(node) == np.int32:
            return 'n{}'.format(node, 1)
        if len(node.vs) == 1:
            return 'n{}'.format(node.id, 1)

        return ','.join([self._to_newick_leaf_bottom_up(self.node_list[v]) for v in node.vs])

    def _to_newick_leaf_top_down(self, node):
        if type(node) == int or type(node) == np.int32:
            return 'n{}:{}'.format(node, 1)
        if len(node.vs) == 1:
            return 'n{}:{}'.format(node.id, 1)

        return ','.join([self._to_newick_leaf_top_down(c) for c in node.children])


class SEAT(AgglomerativeClustering):

    def __init__(self, min_k=1, max_k=10,
                 max_g_ratio=0,
                 affinity='precomputed',
                 strategy='top_down',
                 objective='structure_entropy',
                 n_neighbors=10,
                 corr_cut_off=0.8,
                 ):
        self.min_k = min_k
        self.max_k = max_k
        self.ks = range(min_k, max_k+1)
        self.max_g_ratio = max_g_ratio
        self.affinity = affinity
        self.strategy = strategy
        self.objective = objective
        self.n_neighbors = n_neighbors
        self.corr_cut_off = corr_cut_off

    def get_affinity(self, X):
        if self.affinity == 'precomputed':
            aff_m = X
        elif self.affinity == 'knn_neighbors':
            # X = umap.UMAP().fit_transform(np.matrix(X))
            aff_m = kneighbors_graph(X, self.n_neighbors).toarray()
            aff_m = (aff_m + aff_m.T)/2
            aff_m[np.nonzero(aff_m)] = 1

        elif self.affinity == 'T10':
            # idx = np.argwhere(np.std(X, axis=1) == 0).T[0]
            # X[idx, 0] = X[idx, 0] + 0.0001
            # aff_m = np.corrcoef(X)
            from scipy.stats import spearmanr
            from scipy.spatial.distance import squareform, pdist
            aff_m, _ = spearmanr(X.T)
            aff_m[aff_m < 0] = 0
            dist_m = squareform(pdist(X, metric='euclidean'))
            sigma = 10
            aff_m2 = np.exp(-dist_m*dist_m/(2*sigma*sigma))
            aff_m2[aff_m2 < 0] = 0
            aff_m = (0.8*aff_m + 0.2*aff_m2)/2
            aff_m[aff_m < 0.25] = 0

        elif self.affinity == 'T16':
            # idx = np.argwhere(np.std(X, axis=1) == 0).T[0]
            # X[idx, 0] = X[idx, 0] + 0.0001
            # aff_m = np.corrcoef(X)
            from scipy.stats import spearmanr
            from scipy.spatial.distance import squareform, pdist
            aff_m, _ = spearmanr(X.T)
            aff_m[aff_m < 0] = 0
            dist_m = squareform(pdist(X, metric='euclidean'))
            sigma = 10
            aff_m2 = np.exp(-dist_m*dist_m/(2*sigma*sigma))
            aff_m2[aff_m2 < 0] = 0
            aff_m = (0.8*aff_m + 0.2*aff_m2)/2
            # aff_m[aff_m < 0.25] = 0

        self.affinity_m = aff_m

    def fit(self, X, y=None):

        X = self._validate_data(X, ensure_min_samples=2, estimator=self)

        if self.min_k is not None and self.min_k <= 0:
            raise ValueError("min_k should be an integer greater than 0."
                             " %s was provided." % str(self.min_k))

        if self.max_k is not None and self.max_k <= 2:
            raise ValueError("max_k should be an integer greater than 2."
                             " %s was provided." % str(self.max_k))

        if self.affinity not in ['precomputed', 'knn_neighbors', 'T10', 'T16']:
            raise ValueError("affinity should be precomputed, knn_neighbors, correlation."
                             " %s was provided." % str(self.affinity))

        if self.strategy not in ['bottom_up', 'top_down']:
            raise ValueError("affinity should be bottom_up, top_down."
                             " %s was provided." % str(self.strategy))

        self.get_affinity(X)

        # build the tree

        setree_class = pySETree
        # setree_class = seat_wrapper.SETree

        se_tree = setree_class(self.affinity_m, self.min_k, self.max_k,
                               self.max_g_ratio,
                               self.objective,
                               self.strategy)
        self.se_tree = se_tree
        t1 = time.time()
        Z = se_tree.build_tree()
        t2 = time.time()
        print(Z[:20, :])
        print('build tree time', t2 - t1)
        self.affinity_m = se_tree.affinity_m
        se_tree.cut_tree(Z, self.ks)
        self.vertex_num = se_tree.vertex_num
        self.ks = list(se_tree.ks)
        self.se_scores = se_tree.se_scores
        self.delta_se_scores = se_tree.delta_se_scores
        self.optimal_k = se_tree.optimal_k
        self.labels_ = se_tree.optimal_clusters
        self.Z_ = Z[:, :4]
        self.leaves_list = hierarchy.leaves_list(self.Z_)
        self.order = self._order()
        self.ks_clusters = se_tree.ks_clusters
        self.Z_clusters = se_tree.Z_clusters
        self.submodules = self._get_submodules()
        self.submodule_k = len(se_tree.leafs)

        self.newick = se_tree.to_newick()

        return self

    def _order(self):
        # hierarchy.leaves_list(self.Z_)
        order = [(l, i) for i, l in enumerate(self.leaves_list)]
        order.sort()
        return [i for l, i in order]

    def _get_submodules(self):
        leafs = sorted([(self.order[l.vs[0]], l.id) for l in self.se_tree.leafs])
        order = [(v, i) for i, l in enumerate(leafs) for v in self.se_tree.node_list[l[1]].vs]
        order.sort()
        return [i for n, i in order]

    def oval_embedding(self, a=3, b=2, k=0.2):
        angle = np.array([self.order])*(2*np.pi/len(self.order))
        xcor = a*np.cos(angle)
        ycor = b*np.sqrt(np.exp(k*a*np.cos(angle)))*np.sin(angle)
        plane_coordinate = np.concatenate((xcor, ycor), axis=0).T
        return plane_coordinate



# -*- coding: utf-8 -*-
"""
    src.HE
    ~~~~~~~~~~~

    @Copyright: (c) 2022-07 by Lingxi Chen (chanlingxi@gmail.com).
    @License: LICENSE_NAME, see LICENSE for more details.
"""

import torch
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import seaborn as sns


def find_ab_params(spread, min_dist):
    """Fit a, b params for the differentiable curve used in lower
    dimensional fuzzy simplicial complex construction. We want the
    smooth curve (from a pre-defined family with simple gradient) that
    best matches an offset exponential decay.
    """

    def curve(x, a, b):
        return 1.0 / (1.0 + a * x ** (2 * b))

    xv = np.linspace(0, spread * 3, 300)
    yv = np.zeros(xv.shape)
    yv[xv < min_dist] = 1.0
    yv[xv >= min_dist] = np.exp(-(xv[xv >= min_dist] - min_dist) / spread)
    params, covar = curve_fit(curve, xv, yv)
    return params[0], params[1]


class HierachicalEmbedding(object):

    def __init__(self, n_components=2,
                 affinity='precomputed',
                 n_epochs=None,
                 learning_rate=1e-2,
                 init='spectral',
                 spectral_init_y_idx=1,
                 min_dist=0.1,
                 spread = 1,
                 device='cpu',
                 verbose=True,
                 random_state=None):
        self.n_components = n_components
        self.affinity = affinity
        self.n_epochs = n_epochs
        self.learning_rate = learning_rate
        self.init = init
        self.spectral_init_y_idx = spectral_init_y_idx
        self.min_dist = min_dist
        self.spread = spread
        self.device = device
        self.verbose = verbose
        self.random_state = random_state
        
        self.a, self.b = find_ab_params(self.spread, self.min_dist)
        if self.verbose:
            print('spread = {}, min_dist = {}, a = {}, b = {}'.format(self.spread, self.min_dist, self.a, self.b))
        
    def get_similarity(self, X):
        D = self.get_distance(X)
        S = 1 / (1 + self.a*(D**self.b))
        return S

    def get_distance(self, X):
        M, N = X.shape
        G = torch.mm(X, X.t())
        g = torch.diag(G)
        x = g.repeat(M, 1)
        D = x + x.t() - 2*G
        return D
    
    def replace_submatrix(self, mat, ind1, ind2, mat_replace):
        for i, index in enumerate(ind1):
            mat[index, ind2] = mat_replace[i, :]
    
    def process_affinity(self, X, y):
        if isinstance(y, pd.DataFrame):
            y = y.to_numpy()
        aff_m = X
        self.aff_m_list = []
        self.aff_t_list = []

        for i in range(y.shape[1]):
            tmp_aff_m = np.zeros((self.n_sample, self.n_sample))
            labels = y[:, i]           
            for label in set(labels):
                idx = np.argwhere(labels == label)[:, 0]
                self.replace_submatrix(tmp_aff_m, idx, idx, aff_m[np.ix_(idx, idx)]) 
                
            self.aff_m_list.append(tmp_aff_m)
            self.aff_t_list.append(torch.tensor(tmp_aff_m, dtype=torch.float, device=self.device))
        
    def init_U(self):
        if self.init == 'random':
            generator = torch.Generator(device=self.device)
            if self.random_state is not None:
                generator.manual_seed(self.random_state)
            if self.aff_m_list[0].shape[0] > 50:
                k = 50
                k = self.aff_m_list[0].shape[0]
            else:
                k = self.aff_m_list[0].shape[0]

            U = torch.rand(self.n_sample, k, dtype=torch.float, 
                           device=self.device, generator=generator, requires_grad=True)
            self.inital_U = U.data.cpu().detach().numpy().copy()        
        if self.init == 'spectral':
            from sklearn.manifold import SpectralEmbedding
            self.inital_U = SpectralEmbedding(n_components=self.n_components, random_state=self.random_state,
                                              affinity='precomputed').fit_transform(self.aff_m_list[self.spectral_init_y_idx])
            U = torch.tensor(self.inital_U, dtype=torch.float, device=self.device, requires_grad=True)
        return U
    
    def lap_loss(self, X, Sw):
        return torch.sum(get_distance(X)*Sw)

    def bce_loss(self, club_S, U_S, eph=1e-4):
        return torch.sum(club_S*torch.log(torch.div(club_S, U_S+eph)+eph) + (1-club_S)*torch.log(torch.div(1-club_S, 1-U_S+eph)+eph))

    def fit_transform(self, X, y, thetas=[0.0, 0.2, 0.8]):

        self.n_sample = X.shape[0]
        self.y = y
        if y.shape[1] == 1:
            self.spectral_init_y_idx = 0
        self.thetas = thetas
        self.process_affinity(X, y)

        U = self.init_U()
        
        optimizer = torch.optim.Adam([U], lr=self.learning_rate, weight_decay=0.0001)  # append L2 penalty, 

        loss_list = []
          
        if self.n_epochs is None:
            # For smaller datasets we can use more epochs
            if X.shape[0] <= 10000:
                if self.init == 'spetral':
                    self.n_epochs = 1000
                else:
                    self.n_epochs = 500
            else:
                if self.init == 'spetral':
                    self.n_epochs = 600
                else:
                    self.n_epochs = 200  
        
        
        for i in range(self.n_epochs):
            self.a = 1
            self.b = 1
            U_S = self.get_similarity(U)
    
            info = 'epoch={}\t'.format(i)
        
            loss = 0
            for j in range(0, len(self.aff_t_list)):
                aff_m = self.aff_t_list[j]
                tmp_loss = thetas[j] * self.bce_loss(aff_m, U_S)
                loss = loss + tmp_loss
                info += 'loss_{}={}\t'.format(j, tmp_loss)
            info += 'loss={}\t'.format(loss)
                
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            current_loss = loss.item()
            info += 'update_loss={}\t'.format(current_loss)
            loss_list.append(current_loss)
           
            if i % 100 == 0 and self.verbose:
                print(info)
        if self.verbose:
            print(info)    
        
        self.U = U
        self.loss_list = loss_list
        self.embed_U = U.data.cpu().detach().numpy()[:,:self.n_components]
        return self.embed_U    

    def viz_fit(self, n_row=3, n_col=None, fig_width=5, fn=None):
        if n_col is None:
            n_col = self.y.shape[1]
        point_size = 100.0 / np.sqrt(self.inital_U.shape[0])
        plt.figure(figsize=(fig_width*n_col, fig_width*n_row))
        fig_count = 1
        
        if isinstance(self.y, np.ndarray):
            y = pd.DataFrame(self.y, columns=['col{}'.format(i) for i in range(self.y.shape[1])])
        else:
            y = self.y
        
        args = 'min_dist={},thetas={}'.format(self.min_dist, self.thetas)
        for i in range(y.shape[1]):
            plt.subplot(n_row, n_col, fig_count)
            plt.title('{} init,c={}\n{}'.format(self.init, y.columns[i], args))
            plt.scatter(self.inital_U[:, 0], self.inital_U[:, 1], c=y[y.columns[i]], 
                        s=point_size,
                        cmap=ListedColormap(sns.color_palette('Spectral', len(set(y[y.columns[i]])))))
            fig_count += 1

        for i in range(self.y.shape[1]):
            plt.subplot(n_row, n_col, fig_count)
            plt.title('HE,c={}\n{}'.format(y.columns[i], args))
            plt.scatter(self.embed_U[:, 0], self.embed_U[:, 1], 
                        s=point_size,
                        c=y[y.columns[i]], 
                        cmap=ListedColormap(sns.color_palette('Spectral', len(set(y[y.columns[i]])))))
            fig_count += 1

        plt.subplot(n_row, n_col, fig_count)
        plt.title('Training loss')
        plt.plot(self.loss_list)
        fig_count += 1 
        
        if fn is None:
            plt.show()
        else:
            plt.savefig(fn)
       

   






import setuptools
from pyseat.version import __version__

# with open("README.md", "r") as fh:
#    long_description = fh.read()

setuptools.setup(
    name="pyseat",
    version=__version__,
    author="Lingxi Chen",
    author_email="chanlingxi@gmail.com",
    description="Structure Entropy hierArchy deTection (SEAT) for clustering, ordering, and embedding",
    # long_description=long_description,
    # long_description_content_type="text/markdown",
    url="https://github.com/deepomicslab/seat",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'cycler>=0.11.0',
        'fonttools>=4.37.1',
        'joblib>=1.1.0',
        'kiwisolver>=1.4.4',
        'kmeans1d>=0.3.1',
        'matplotlib>=3.5.3',
        'networkx>=2.8.6',
        'numpy>=1.23.3',
        'packaging>=21.3',
        'pandas>=1.4.4',
        'Pillow>=9.2.0',
        'pyparsing>=3.0.9',
        'python-dateutil>=2.8.2',
        'pytz>=2022.2.1',
        'scikit-learn>=1.1.2',
        'scipy>=1.9.1',
        'seaborn>=0.12.0',
        'six>=1.16.0',
        'sklearn>=0.0',
        'threadpoolctl>=3.1.0',
        'torch>=1.12.1',
        'typing_extensions>=4.3.0'
    ]
)



# _*_ coding: utf-8 _*_
"""
Time:     2021/8/27 21:32
Author:   WANG Bingchen
Version:  V 0.1
File:     distance.py
Describe: 
"""

import numba
import numpy as np


@numba.njit()
def adjustment(x, k, a):
    return np.exp(k*a*np.cos(x))


@numba.njit()
def adjustmentGrad(x, k, a):
    return -k*a*np.exp(k*a*np.cos(x))*np.sin(x)



@numba.njit(fastmath=True)
def distanceInOval(x, y, a=3, b=2, k=0.2):
    """

    :param x: high-dimension embedding of cell A
    :param y: high-dimension embedding of cell B
    :param a: major axis length
    :param b: minor axie length
    :param k: Deformation parameter

    :return: distance between cell A and B in oval whose function is

                    x^2/a^2 + y^2/(t(x)*b^2) = 1

                    where t(x) = exp(kx)
    """

    result = 0.0

    x = x % (2*np.pi)
    y = y % (2*np.pi)
    for i in range(x.shape[0]):
        result += (a*np.cos(x[i]) - a*np.cos(y[i]))**2 + (b*np.sqrt(adjustment(x[i], k, a))*np.sin(x[i]) - b*np.sqrt(adjustment(y[i], k, a))*np.sin(y[i]))**2
    d = np.sqrt(result)
    grad1 = -(a**2)*(np.cos(x) - np.cos(y))*np.sin(x)
    grad2 = (b**2)*(np.sqrt(adjustment(x, k, a))*np.sin(x) - np.sqrt(adjustment(y, k, a))*np.sin(y))
    grad3 = adjustmentGrad(x, k, a)*np.sin(x)/(2*np.sqrt(adjustment(x, k, a))) + np.cos(x)*np.sqrt(adjustment(x, k, a))
    grad = grad1 + grad2 * grad3

    grad = grad/(d + 1e-6)

    return d, grad



# _*_ coding: utf-8 _*_
"""
Time:     2021/8/27 21:33
Author:   WANG Bingchen
Version:  V 0.1
File:     cell_circle.py
Describe: 
"""

import numpy as np
import umap
from .distance import distanceInOval





def angle2plane_coordinate(angle, a, b, k):
    xcor = a*np.cos(angle)
    ycor = b*np.sqrt(np.exp(k*a*np.cos(angle)))*np.sin(angle)

    plane_coordinate = np.concatenate((xcor, ycor), axis=1)
    return plane_coordinate





class CellCircle():

    def __init__(self, n_neighbors=20, min_dist=0.01):
        self.n_neighbors = n_neighbors
        self.min_dist = min_dist
        self.n_components = 1



    def cal_cell_circle(self, data, a=3, b=2, k=0.2):
        """

        :param data: input data array(n_features, n_cells)
        :param a: major axis length
        :param b: minor axie length
        :param k: Deformation parameter
        :return: {
                    angle: embeddings as angles(Unit: rad),
                    plane_embedding: embeddings in plane coordinate
                }
        """
        m, n = data.shape
        print("%d cells, %d features" % (n, m))
        data = data.transpose()

        mapper = umap.UMAP(n_components=self.n_components, n_neighbors=self.n_neighbors, min_dist=self.min_dist,
                           output_metric=distanceInOval, output_metric_kwds={"a": a, "b": b, "k": k}).fit(data)

        angle = mapper.transform(data)
        angle = angle % (2 * np.pi)

        plane_embedding = angle2plane_coordinate(angle, a=a, b=b, k=k)

        return {
            "angle": angle,
            "plane_embedding": plane_embedding
        }